# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006 Fluendo, S.A. (www.fluendo.com).
# Copyright 2006, Frank Scholz <coherence@beebits.net>
from coherence.upnp.core.xml_constants import UPNP_SERVICE_NS

import os

import time
import urllib2
from coherence.upnp.core import action
from coherence.upnp.core import event
from coherence.upnp.core import variable

from coherence.upnp.core import utils
from coherence.upnp.core.soap_proxy import SOAPProxy
from coherence.upnp.core.soap_service import errorCode
from coherence.upnp.core.event import EventSubscriptionServer

from lxml import etree

from twisted.web import static
from twisted.internet import defer, reactor
from twisted.python import failure, util
from twisted.internet import task

import coherence.extern.louie as louie

from coherence import log

global subscribers
subscribers = dict()

NS_UPNP_ORG_EVENT_1_0 = 'urn:schemas-upnp-org:event-1-0'


def subscribe(service):
  subscribers[service.get_sid()] = service


def unsubscribe(service):
  subscribers.pop(service.get_sid(), None)


class Service(log.Loggable):
    logCategory = 'service_client'

    def __init__(self, service_type, service_id, location, control_url,
                 event_sub_url, presentation_url, scpd_url, device):
        log.Loggable.__init__(self)

        self.service_type = service_type
        self.detection_completed = False
        self.id = service_id
        self.control_url = control_url
        self.event_sub_url = event_sub_url
        self.presentation_url = presentation_url
        self.scpd_url = scpd_url
        self.device = device
        self._actions = {}
        self._variables = {0: {}}
        self._var_subscribers = {}
        self.subscription_id = None
        self.timeout = 0

        self.event_connection = None
        self.last_time_updated = None

        self.client = None

        parsed = urllib2.urlparse.urlparse(location)
        self.url_base = "%s://%s" % (parsed[0], parsed[1])

        self.parse_actions()
        self.info("%s %s %s initialized", self.device.friendly_name, self.service_type, self.id)

    def as_tuples(self):
        r = []

        def append(name, attribute):
            try:
                if isinstance(attribute, tuple):
                    if callable(attribute[0]):
                        v1 = attribute[0]()
                    elif hasattr(self, attribute[0]):
                        v1 = getattr(self, attribute[0])
                    else:
                        v1 = attribute[0]
                    if v1 in [None, 'None']:
                        return
                    if callable(attribute[1]):
                        v2 = attribute[1]()
                    elif hasattr(self, attribute[1]):
                        v2 = getattr(self, attribute[1])
                    else:
                        v2 = attribute[1]
                    if v2 in [None, 'None']:
                        return
                    if len(attribute) > 2:
                        r.append((name, (v1, v2, attribute[2])))
                    else:
                        r.append((name, (v1, v2)))
                    return
                elif callable(attribute):
                    v = attribute()
                elif hasattr(self, attribute):
                    v = getattr(self, attribute)
                else:
                    v = attribute
                if v not in [None, 'None']:
                    r.append((name, v))
            except:
                import traceback
                self.debug(traceback.format_exc())

        r.append(('Location', (self.device.get_location(), self.device.get_location())))
        append('URL base', self.device.get_urlbase)
        r.append(('UDN', self.device.get_id()))
        r.append(('Type', self.service_type))
        r.append(('ID', self.id))
        append('Service Description URL', (self.scpd_url, lambda: self.device.make_fullyqualified(self.scpd_url)))
        append('Control URL', (self.control_url, lambda: self.device.make_fullyqualified(self.control_url), False))
        append('Event Subscription URL', (self.event_sub_url, lambda: self.device.make_fullyqualified(self.event_sub_url), False))

        return r

    def as_dict(self):
        d = {'type': self.service_type}
        d['actions'] = [a.as_dict() for a in self._actions.values()]
        return d

    def __repr__(self):
        return "Service %s %s" % (self.service_type, self.id)
    #def __del__(self):
    #    print "Service deleted"
    #    pass

    def _get_client(self, name):
        url = self.get_control_url()
        namespace = self.get_type()
        action = "%s#%s" % (namespace, name)
        client = SOAPProxy(url, namespace=("u", namespace), soapaction=action)
        return client

    def remove(self):
        self.info("removal of  %s %s %s", self.device.friendly_name, self.service_type, self.id)
        try:
            self.renew_subscription_call.cancel()
        except:
            pass
        if self.event_connection != None:
            self.event_connection.teardown()
        if self.subscription_id != None:
            self.unsubscribe()
        for name, action in self._actions.items():
            self.debug("remove %s %s", name, action)
            del self._actions[name]
            del action
        for instance, variables in self._variables.items():
            for name, variable in variables.items():
                del variables[name]
                del variable
            if variables.has_key(instance):
                del variables[instance]
            del variables
        del self

    def get_device(self):
        return self.device

    def get_type(self):
        return self.service_type

    def set_timeout(self, timeout):
        self.info("set timout for %s/%s to %d", self.device.friendly_name, self.service_type, int(timeout))
        self.timeout = timeout
        try:
            self.renew_subscription_call.reset(int(self.timeout) - 30)
            self.info("reset renew subscription call for %s/%s to %d", self.device.friendly_name, self.service_type, int(self.timeout) - 30)
        except:
            self.renew_subscription_call = reactor.callLater(int(self.timeout) - 30, self.renew_subscription)
            self.info("starting renew subscription call for %s/%s to %d", self.device.friendly_name, self.service_type, int(self.timeout) - 30)

    def get_timeout(self):
        return self.timeout

    def get_id(self):
        return self.id

    def get_sid(self):
        return self.subscription_id

    def set_sid(self, sid):
        self.info("set subscription id for %s/%s to %s", self.device.friendly_name, self.service_type, sid)
        self.subscription_id = sid
        if sid is not None:
            subscribe(self)
            self.debug("add subscription for %s", self.id)

    def get_actions(self):
        return self._actions

    def get_scpdXML(self):
        return self.scpdXML

    def get_action(self, name):
        try:
            return self._actions[name]
        except KeyError:
            return None  # not implemented

    def get_state_variables(self, instance):
        return self._variables.get(int(instance))

    def get_state_variable(self, name, instance=0):
        return self._variables.get(int(instance)).get(name)

    def get_control_url(self):
        return self.device.make_fullyqualified(self.control_url)

    def get_event_sub_url(self):
        return self.device.make_fullyqualified(self.event_sub_url)

    def get_presentation_url(self):
        return self.device.make_fullyqualified(self.presentation_url)

    def get_scpd_url(self):
        return self.device.make_fullyqualified(self.scpd_url)

    def get_base_url(self):
        return self.device.make_fullyqualified('.')

    def subscribe(self):
        self.debug("subscribe %s", self.id)
        event.subscribe(self)
        #global subscribers
        #subscribers[self.get_sid()] = self

    def unsubscribe(self):

        def remove_it(r, sid):
            self.debug("remove subscription for %s", self.id)
            unsubscribe(self)
            self.subscription_id = None
            #global subscribers
            #if subscribers.has_key(sid):
            #    del subscribers[sid]

        self.debug("unsubscribe %s", self.id)
        d = event.unsubscribe(self)
        d.addCallback(remove_it, self.get_sid())
        return d

    def subscribe_for_variable(self, var_name, instance=0, callback=None, signal=False):
      variable = self.get_state_variable(var_name)
      if variable:
        if callback is not None:
          if signal:
            callback(variable)
            louie.connect(callback, signal='Coherence.UPnP.StateVariable.%s.changed' % var_name, sender=self)
          else:
            variable.subscribe(callback)

    def renew_subscription(self):
      self.info("renew_subscription")
      event.subscribe(self)

    def process_event(self, event):
      self.info("process event %r %r", self, event)
      for var_name, var_value in event.items():
        if var_name == 'LastChange':
            self.info("we have a LastChange event")
            self.get_state_variable(var_name, 0).update(var_value)
            tree = etree.fromstring(var_value)
            namespace_uri, tag = tree.tag[1:].split("}", 1)
            for instance in tree.findall('{%s}InstanceID' % namespace_uri):
                instance_id = instance.attrib['val']
                self.info("instance_id %r %r", instance, instance_id)
                for var in instance.getchildren():
                    self.info("var %r", var)
                    namespace_uri, tag = var.tag[1:].split("}", 1)
                    self.info("%r %r %r", namespace_uri, tag, var.attrib['val'])
                    self.get_state_variable(tag, instance_id).update(var.attrib['val'])
                    self.info("updated var %r", var)
                    if len(var.attrib) > 1:
                        self.info("Extended StateVariable %s - %r", var.tag, var.attrib)
                        if var.attrib.has_key('channel') and var.attrib['channel'] != 'Master':
                            # TODO handle attributes that them selves have multiple instances
                            self.info("Skipping update to %s its not for master channel %s", var.tag, var.attrib)
                            pass
                        else:
                            if not self.get_state_variables(instance_id):
                                # TODO Create instance ?
                                self.error("%r update failed (not self.get_state_variables(instance_id)) %r", self, instance_id)
                            elif not self.get_state_variables(instance_id).has_key(tag):
                                # TODO Create instance StateVariable?
                                # SONOS stuff
                                self.error("%r update failed (not self.get_state_variables(instance_id).has_key(tag)) %r", self, tag)
                            else:
                                val = None
                                if var.attrib.has_key('val'):
                                    val = var.attrib['val']
                                #self.debug("%r update %r %r %r", self,namespace_uri, tag, var.attrib['val'])
                                self.get_state_variable(tag, instance_id).update(var.attrib['val'])
                                self.debug("updated 'attributed' var %r", var)
            louie.send('Coherence.UPnP.DeviceClient.Service.Event.processed', None, self, (var_name, var_value, event.raw))
        else:
            self.get_state_variable(var_name, 0).update(var_value)
            louie.send('Coherence.UPnP.DeviceClient.Service.Event.processed', None, self, (var_name, var_value, event.raw))
      if self.last_time_updated is None:
        # The clients (e.g. media_server_client) check for last time to detect whether service detection is complete
        # so we need to set it here and now to avoid a potential race condition
        self.last_time_updated = time.time()
        louie.send('Coherence.UPnP.DeviceClient.Service.notified', sender=self.device, service=self)
        self.info("send signal Coherence.UPnP.DeviceClient.Service.notified for %r", self)
      self.last_time_updated = time.time()

    def parse_actions(self):

        def gotPage(x):
            self.scpdXML, headers = x
            try:
              tree = etree.fromstring(self.scpdXML)
            except Exception as e:
              self.warning("Invalid service description received from %r: %r", self.get_scpd_url(), e)
              return
            ns = UPNP_SERVICE_NS

            for action_node in tree.findall('.//{%s}action' % ns):
                name = action_node.findtext('{%s}name' % ns)
                arguments = []
                for argument in action_node.findall('.//{%s}argument' % ns):
                    arg_name = argument.findtext('{%s}name' % ns)
                    arg_direction = argument.findtext('{%s}direction' % ns)
                    arg_state_var = argument.findtext('{%s}relatedStateVariable' % ns)
                    arguments.append(action.Argument(arg_name, arg_direction,
                                                     arg_state_var))
                self._actions[name] = action.Action(self, name, 'n/a', arguments)

            for var_node in tree.findall('.//{%s}stateVariable' % ns):
                send_events = var_node.attrib.get('sendEvents', 'yes')
                name = var_node.findtext('{%s}name' % ns)
                data_type = var_node.findtext('{%s}dataType' % ns)
                values = []
                """ we need to ignore this, as there we don't get there our
                    {urn:schemas-beebits-net:service-1-0}X_withVendorDefines
                    attibute there
                """
                for allowed in var_node.findall('.//{%s}allowedValue' % ns):
                    values.append(allowed.text)
                instance = 0
                self._variables.get(instance)[name] = variable.StateVariable(self, name,
                                                               'n/a',
                                                               instance, send_events,
                                                               data_type, values)
                """ we need to do this here, as there we don't get there our
                    {urn:schemas-beebits-net:service-1-0}X_withVendorDefines
                    attibute there
                """
                self._variables.get(instance)[name].has_vendor_values = True

            #print 'service parse:', self, self.device
            self.detection_completed = True
            louie.send('Coherence.UPnP.Service.detection_completed', sender=self.device, device=self.device)
            self.info("send signal Coherence.UPnP.Service.detection_completed for %r", self)
            """
            if (self.last_time_updated == None):
                if( self.id.endswith('AVTransport') or
                    self.id.endswith('RenderingControl')):
                    louie.send('Coherence.UPnP.DeviceClient.Service.notified', sender=self.device, service=self)
                    self.last_time_updated = time.time()
            """

        def gotError(failure, url):
            self.warning('error requesting %s', url)
            self.info('failure %s', failure)
            louie.send('Coherence.UPnP.Service.detection_failed', self.device, device=self.device)

        utils.getPage(self.get_scpd_url()).addCallbacks(gotPage, gotError, None, None, [self.get_scpd_url()], None)

moderated_variables = \
        {'urn:schemas-upnp-org:service:AVTransport:2':
            ['LastChange'],
         'urn:schemas-upnp-org:service:AVTransport:1':
            ['LastChange'],
         'urn:schemas-upnp-org:service:ContentDirectory:2':
            ['SystemUpdateID', 'ContainerUpdateIDs'],
         'urn:schemas-upnp-org:service:ContentDirectory:1':
            ['SystemUpdateID', 'ContainerUpdateIDs'],
         'urn:schemas-upnp-org:service:RenderingControl:2':
            ['LastChange'],
         'urn:schemas-upnp-org:service:RenderingControl:1':
            ['LastChange'],
         'urn:schemas-upnp-org:service:ScheduledRecording:1':
            ['LastChange'],
        }


class ServiceServer(log.Loggable):
    logCategory = 'service_server'

    def __init__(self, id, version, backend):
      log.Loggable.__init__(self)
      self.id = id
      self.version = version
      self.backend = backend
      if getattr(self, "namespace", None) is None:
        self.namespace = 'schemas-upnp-org'
      if getattr(self, "id_namespace", None) is None:
        self.id_namespace = 'upnp-org'

      self.service_type = 'urn:%s:service:%s:%d' % (self.namespace, id, int(self.version))

      self.scpd_url = 'scpd.xml'
      self.control_url = 'control'
      self.subscription_url = 'subscribe'
      self.event_metadata = ''
      if id == 'AVTransport':
        self.event_metadata = 'urn:schemas-upnp-org:metadata-1-0/AVT/'
      if id == 'RenderingControl':
        self.event_metadata = 'urn:schemas-upnp-org:metadata-1-0/RCS/'
      if id == 'ScheduledRecording':
        self.event_metadata = 'urn:schemas-upnp-org:av:srs-event'

      self._actions = {}
      self._variables = {0: {}}
      self._subscribers = {}

      self._pending_notifications = {}

      self.last_change = None
      self.init_var_and_actions()

      try:
        if 'LastChange' in moderated_variables[self.service_type]:
          self.last_change = self._variables[0]['LastChange']
      except:
        pass

      self.putChild(self.subscription_url, EventSubscriptionServer(self))

      self.check_subscribers_loop = task.LoopingCall(self.check_subscribers)
      self.check_subscribers_loop.start(120.0, now=False)

      self.check_moderated_loop = None
      if moderated_variables.has_key(self.service_type):
        self.check_moderated_loop = task.LoopingCall(self.check_moderated_variables)
        self.check_moderated_loop.start(0.5, now=False)

    def _release(self):
        for p in self._pending_notifications.values():
            p.disconnect()
        self._pending_notifications = {}

    def get_action(self, action_name):
        try:
            return self._actions[action_name]
        except KeyError:
            return None  # not implemented

    def get_actions(self):
        return self._actions

    def get_variables(self):
        return self._variables

    def get_subscribers(self):
        return self._subscribers

    def rm_notification(self, result, d):
        del self._pending_notifications[d]

    def new_subscriber(self, subscriber):
        notify = []
        for vdict in self._variables.values():
          notify += [v for v in vdict.values() if v.send_events == True]

        self.info("new_subscriber %s %s", subscriber, notify)
        if len(notify) <= 0:
          return

        root = etree.Element('{%s}propertyset' % NS_UPNP_ORG_EVENT_1_0, nsmap={'e': NS_UPNP_ORG_EVENT_1_0})
        evented_variables = 0
        for n in notify:
          e = etree.SubElement(root, '{%s}property' % NS_UPNP_ORG_EVENT_1_0)
          if n.name == 'LastChange':
            if subscriber['seq'] == 0:
                text = self.build_last_change_event(n.instance, force=True)
            else:
                text = self.build_last_change_event(n.instance)
            if text is not None:
                etree.SubElement(e, n.name).text = text
                evented_variables += 1
          else:
            etree.SubElement(e, n.name).text = str(n.value)
            evented_variables += 1

        if evented_variables > 0:
          xml = etree.tostring(root, encoding='utf-8', pretty_print=True)
          d, p = event.send_notification(subscriber, xml)
          self._pending_notifications[d] = p
          d.addBoth(self.rm_notification, d)
        self._subscribers[subscriber['sid']] = subscriber

    def get_id(self):
        return self.id

    def get_type(self):
        return self.service_type

    def create_new_instance(self, instance):
        self._variables[instance] = {}
        for v in self._variables[0].values():
            self._variables[instance][v.name] = variable.StateVariable(v.service,
                                                                        v.name,
                                                                        v.implementation,
                                                                        instance,
                                                                        v.send_events,
                                                                        v.data_type,
                                                                        v.allowed_values)
            self._variables[instance][v.name].has_vendor_values = v.has_vendor_values
            self._variables[instance][v.name].default_value = v.default_value
            #self._variables[instance][v.name].value = v.default_value # FIXME
            self._variables[instance][v.name].old_value = v.old_value
            self._variables[instance][v.name].value = v.value
            self._variables[instance][v.name].dependant_variable = v.dependant_variable

    def remove_instance(self, instance):
        if instance == 0:
            return
        del(self._variables[instance])

    def set_variable(self, instance, variable_name, value, default=False):

        def process_value(result):
            variable.update(result)
            if default:
                variable.default_value = variable.value
            if variable.send_events and not variable.moderated and len(self._subscribers) > 0:
                xml = self.build_single_notification(instance, variable_name, variable.value)
                for s in self._subscribers.values():
                    d, p = event.send_notification(s, xml)
                    self._pending_notifications[d] = p
                    d.addBoth(self.rm_notification, d)
        try:
            variable = self._variables[int(instance)][variable_name]
            if isinstance(value, defer.Deferred):
                value.addCallback(process_value)
            else:
                process_value(value)
        except:
            pass

    def get_variable(self, variable_name, instance=0):
        try:
            return self._variables[int(instance)][variable_name]
        except:
            return None

    def build_single_notification(self, instance, variable_name, value):
        root = etree.Element('{%s}propertyset' % NS_UPNP_ORG_EVENT_1_0, nsmap={'e': NS_UPNP_ORG_EVENT_1_0})
        e = etree.SubElement(root, '{%s}property' % NS_UPNP_ORG_EVENT_1_0)
        etree.SubElement(e, variable_name).text = str(value)
        return etree.tostring(root, encoding='utf-8', pretty_print=True)

    def build_last_change_event(self, instance=0, force=False):
        got_one = False
        root = etree.Element('Event', nsmap={None: self.event_metadata})
        for instance, vdict in self._variables.items():
          e = etree.SubElement(root, 'InstanceID')
          e.attrib['val'] = str(instance)
          for variable in vdict.values():
            if variable.name != 'LastChange' and variable.name[0:11] != 'A_ARG_TYPE_' and not variable.never_evented:
              if variable.updated or force:
                s = etree.SubElement(e, variable.name)
                s.attrib['val'] = str(variable.value)
                variable.updated = False
                got_one = True
                if variable.dependant_variable is not None:
                  dependants = variable.dependant_variable.get_allowed_values()
                  if dependants is not None and len(dependants) > 0:
                    s.attrib['channel'] = dependants[0]
        if got_one:
            return etree.tostring(root, encoding='utf-8', pretty_print=True)
        else:
            return None

    def propagate_notification(self, notify):
        if len(self._subscribers) <= 0:
          return
        if len(notify) <= 0:
          return

        root = etree.Element('{%s}propertyset' % NS_UPNP_ORG_EVENT_1_0, nsmap={'e': NS_UPNP_ORG_EVENT_1_0})

        if isinstance(notify, variable.StateVariable):
          notify = [notify, ]

        evented_variables = 0
        for n in notify:
          e = etree.SubElement(root, '{%s}property' % NS_UPNP_ORG_EVENT_1_0)
          if n.name == 'LastChange':
            text = self.build_last_change_event(instance=n.instance)
            if text is not None:
              etree.SubElement(e, n.name).text = text
              evented_variables += 1
          else:
            s = etree.SubElement(e, n.name).text = str(n.value)
            evented_variables += 1
            if n.dependant_variable is not None:
              dependants = n.dependant_variable.get_allowed_values()
              if dependants is not None and len(dependants) > 0:
                s.attrib['channel'] = dependants[0]

        if evented_variables == 0:
            return
        xml = etree.tostring(root, encoding='utf-8', pretty_print=True)

        for s in self._subscribers.values():
            d, p = event.send_notification(s, xml)
            self._pending_notifications[d] = p
            d.addBoth(self.rm_notification, d)

    def check_subscribers(self):
        for s in self._subscribers.values():
            timeout = 86400
            if s['timeout'].startswith('Second-'):
                timeout = int(s['timeout'][len('Second-'):])
            if time.time() > s['created'] + timeout:
                del s

    def check_moderated_variables(self):
        #print "check_moderated for %s" % self.id
        #print self._subscribers
        if len(self._subscribers) <= 0:
            return
        variables = moderated_variables[self.get_type()]
        notify = []
        for v in variables:
            #print self._variables[0][v].name, self._variables[0][v].updated
            for vdict in self._variables.values():
                if vdict[v].updated == True:
                    vdict[v].updated = False
                    notify.append(vdict[v])
        self.propagate_notification(notify)

    def is_variable_moderated(self, name):
        try:
            variables = moderated_variables[self.get_type()]
            if name in variables:
                return True
        except:
            pass
        return False

    def simulate_notification(self):
        print "simulate_notification for", self.id
        self.set_variable(0, 'CurrentConnectionIDs', '0')

    def get_scpdXML(self):
        if not hasattr(self, 'scpdXML') or self.scpdXML == None:
            self.scpdXML = scpdXML(self)
            self.scpdXML = self.scpdXML.build_xml()
        return self.scpdXML

    def register_vendor_variable(self, name, implementation='optional',
                                      instance=0, evented='no',
                                      data_type='string',
                                      dependant_variable=None,
                                      default_value=None,
                                      allowed_values=None, has_vendor_values=False, allowed_value_range=None,
                                      moderated=False):
        """
        enables a backend to add an own, vendor defined, StateVariable to the service

        @ivar name: the name of the new StateVariable
        @ivar implementation: either 'optional' or 'required'
        @ivar instance: the instance number of the service that variable should be assigned to, usually '0'
        @ivar evented: boolean, or the special keyword 'never' if the variable doesn't show up in a LastChange event too
        @ivar data_type: 'string','boolean','bin.base64' or various number formats
        @ivar dependant_variable: the name of another StateVariable that depends on this one
        @ivar default_value: the value this StateVariable should have by default when created
                             for another instance of in the service
        @ivar allowed_values: a C{list} of values this StateVariable can have
        @ivar has_vendor_values: boolean if there are values outside the allowed_values list too
        @ivar allowed_value_range: a C{dict} of 'minimum','maximum' and 'step' values
        @ivar moderated: boolean, True if this StateVariable should only be evented via a LastChange event

        """

        # FIXME
        # we should raise an Exception when there as a StateVariable with that name already

        if evented == 'never':
            send_events = 'no'
        else:
            send_events = evented
        new_variable = variable.StateVariable(self, name, implementation, instance, send_events,
                                                   data_type, allowed_values)
        if default_value == None:
            new_variable.default_value = ''
        else:
            new_variable.default_value = new_variable.old_value = new_variable.value = default_value

        new_variable.dependant_variable = dependant_variable
        new_variable.has_vendor_values = has_vendor_values
        new_variable.allowed_value_range = allowed_value_range
        new_variable.moderated = moderated
        if evented == 'never':
            new_variable.never_evented = True
        self._variables.get(instance)[name] = new_variable
        return new_variable

    def register_vendor_action(self, name, implementation, arguments=None, needs_callback=True):
        """
        enables a backend to add an own, vendor defined, Action to the service

        @ivar name: the name of the new Action
        @ivar implementation: either 'optional' or 'required'
        @ivar arguments: a C{list} if argument C{tuples},
                         like (name,direction,relatedStateVariable)
        @ivar needs_callback: this Action needs a method in the backend or service class
        """
        # FIXME
        # we should raise an Exception when there as an Action with that name already
        # we should raise an Exception when there is no related StateVariable for an Argument

        """ check for action in backend """
        callback = getattr(self.backend, "upnp_%s" % name, None)

        if callback is None:
          """ check for action in ServiceServer """
          callback = getattr(self, "upnp_%s" % name, None)

        if needs_callback and callback is None:
          """ we have one or more 'A_ARG_TYPE_' variables
              issue a warning for now
          """
          if implementation == 'optional':
            self.info('%s has a missing callback for %s action %s, action disabled', self.id, implementation, name)
            return
          else:
            if (hasattr(self, 'implementation') and self.implementation == 'required') or not hasattr(self, 'implementation'):
              self.warning('%s has a missing callback for %s action %s, service disabled', self.id, implementation, name)
            raise LookupError("missing callback")

        arguments_list = []
        for argument in arguments:
            arguments_list.append(action.Argument(argument[0], argument[1].lower(), argument[2]))

        new_action = action.Action(self, name, implementation, arguments_list)
        self._actions[name] = new_action
        if callback is not None:
            new_action.set_callback(callback)
            self.info('Add callback %s for %s/%s', callback, self.id, name)
        return new_action

    def init_var_and_actions(self):
        desc_file = util.sibpath(__file__, os.path.join('xml-service-descriptions', '%s%d.xml' % (self.id, int(self.version))))
        tree = etree.parse(desc_file)

        for action_node in tree.findall('.//action'):
            name = action_node.findtext('name')
            implementation = 'required'
            needs_callback = False
            if action_node.attrib.get('{urn:schemas-beebits-net:service-1-0}X_needs_backend', None) is not None:
              needs_callback = True
            if action_node.find('Optional') is not None:
                implementation = 'optional'
                if action_node.find('Optional').attrib.get('{urn:schemas-beebits-net:service-1-0}X_needs_backend', None) is not None \
                   or action_node.attrib.get('{urn:schemas-beebits-net:service-1-0}X_needs_backend', None) is not None:
                  needs_callback = True

            arguments = []
            for argument in action_node.findall('.//argument'):
                arg_name = argument.findtext('name')
                arg_direction = argument.findtext('direction')
                arg_state_var = argument.findtext('relatedStateVariable')
                arguments.append(action.Argument(arg_name, arg_direction, arg_state_var))
                if arg_state_var[0:11] == 'A_ARG_TYPE_' and arg_direction == 'out':
                  needs_callback = True

            """ check for action in backend """
            callback = getattr(self.backend, "upnp_%s" % name, None)

            if callback is None:
                """ check for action in ServiceServer """
                callback = getattr(self, "upnp_%s" % name, None)

            if needs_callback and callback is None:
                """ we have one or more 'A_ARG_TYPE_' variables
                    issue a warning for now
                """
                if implementation == 'optional':
                  self.info('%s has a missing callback for %s action %s, action disabled', self.id, implementation, name)
                  continue
                else:
                  if (hasattr(self, 'implementation') and self.implementation == 'required') or not hasattr(self, 'implementation'):
                      self.warning('%s has a missing callback for %s action %s, service disabled', self.id, implementation, name)
                  raise LookupError, "missing callback"

            new_action = action.Action(self, name, implementation, arguments)
            self._actions[name] = new_action
            if callback is not None:
              new_action.set_callback(callback)
              self.info('Add callback %s for %s/%s', callback, self.id, name)

        backend_vendor_value_defaults = getattr(self.backend, "vendor_value_defaults", None)
        service_value_defaults = None
        if backend_vendor_value_defaults:
            service_value_defaults = backend_vendor_value_defaults.get(self.id, None)

        backend_vendor_range_defaults = getattr(self.backend, "vendor_range_defaults", None)
        service_range_defaults = None
        if backend_vendor_range_defaults:
            service_range_defaults = backend_vendor_range_defaults.get(self.id)

        for var_node in tree.findall('.//stateVariable'):
            instance = 0
            name = var_node.findtext('name')
            implementation = 'required'
            if action_node.find('Optional') is not None:
                implementation = 'optional'

            send_events = var_node.findtext('sendEventsAttribute')
            data_type = var_node.findtext('dataType')
            values = []
            for allowed in var_node.findall('.//allowedValue'):
                values.append(allowed.text)
            self._variables.get(instance)[name] = variable.StateVariable(self, name, implementation,
                                                                         instance, send_events,
                                                                         data_type, values)

            dependant_variable = var_node.findtext('{urn:schemas-beebits-net:service-1-0}X_dependantVariable')
            if dependant_variable:
                self._variables.get(instance)[name].dependant_variable = dependant_variable
            default_value = var_node.findtext('defaultValue')
            if default_value:
                self._variables.get(instance)[name].set_default_value(default_value)
            if var_node.find('sendEventsAttribute') is not None:
                never_evented = var_node.find('sendEventsAttribute').attrib.get(
                                        '{urn:schemas-beebits-net:service-1-0}X_no_means_never',
                                        None)
                if never_evented is not None:
                    self._variables.get(instance)[name].set_never_evented(never_evented)

            allowed_value_list = var_node.find('allowedValueList')
            if allowed_value_list != None:
                vendor_values = allowed_value_list.attrib.get(
                                    '{urn:schemas-beebits-net:service-1-0}X_withVendorDefines',
                                    None)
                if service_value_defaults:
                    variable_value_defaults = service_value_defaults.get(name, None)
                    if variable_value_defaults:
                        self.info("overwriting %s default value with %s", name,
                                                               variable_value_defaults)
                        self._variables.get(instance)[name].set_allowed_values(variable_value_defaults)

                if vendor_values is not None:
                    self._variables.get(instance)[name].has_vendor_values = True

            allowed_value_range = var_node.find('allowedValueRange')
            if allowed_value_range:
              vendor_values = allowed_value_range.attrib.get('{urn:schemas-beebits-net:service-1-0}X_withVendorDefines', None)
              range = {}
              for e in list(allowed_value_range):
                  range[e.tag] = e.text
                  if vendor_values != None:
                      if service_range_defaults:
                          variable_range_defaults = service_range_defaults.get(name)
                          if(variable_range_defaults != None and
                              variable_range_defaults.get(e.tag) != None):
                              self.info("overwriting %s attribute %s with %s", name,
                                                             e.tag, str(variable_range_defaults[e.tag]))
                              range[e.tag] = variable_range_defaults[e.tag]
                          elif e.text == None:
                              self.info("missing vendor definition for %s, attribute %s", name, e.tag)
              self._variables.get(instance)[name].set_allowed_value_range(**range)
              if vendor_values is not None:
                  self._variables.get(instance)[name].has_vendor_values = True
            elif service_range_defaults:
                variable_range_defaults = service_range_defaults.get(name)
                if variable_range_defaults is not None:
                    self._variables.get(instance)[name].set_allowed_value_range(**variable_range_defaults)
                    self._variables.get(instance)[name].has_vendor_values = True

        for v in self._variables.get(0).values():
          if isinstance(v.dependant_variable, str):
            v.dependant_variable = self._variables.get(instance).get(v.dependant_variable)


class scpdXML(static.Data):

    def __init__(self, server, control=None):
        self.service_server = server
        self.control = control
        static.Data.__init__(self, None, 'text/xml')

    def render(self, request):
        if self.data is None:
            self.data = self.build_xml()
        return static.Data.render(self, request)

    def build_xml(self):
      root = etree.Element('scpd', nsmap={None: 'urn:schemas-upnp-org:service-1-0'})
      e = etree.SubElement(root, 'specVersion')
      etree.SubElement(e, 'major').text = '1'
      etree.SubElement(e, 'minor').text = '0'

      e = etree.SubElement(root, 'actionList')
      for action in self.service_server._actions.values():
        s = etree.SubElement(e, 'action')
        etree.SubElement(s, 'name').text = action.get_name()
        al = etree.SubElement(s, 'argumentList')
        for argument in action.get_arguments_list():
          a = etree.SubElement(al, 'argument')
          etree.SubElement(a, 'name').text = argument.get_name()
          etree.SubElement(a, 'direction').text = argument.get_direction()
          etree.SubElement(a, 'relatedStateVariable').text = argument.get_state_variable()

      e = etree.SubElement(root, 'serviceStateTable')
      for var in self.service_server._variables[0].values():
        s = etree.SubElement(e, 'stateVariable')
        if var.send_events:
          s.attrib['sendEvents'] = 'yes'
        else:
          s.attrib['sendEvents'] = 'no'
        etree.SubElement(s, 'name').text = var.name
        etree.SubElement(s, 'dataType').text = var.data_type
        if not var.has_vendor_values and len(var.allowed_values):
          v = etree.SubElement(s, 'allowedValueList')
          for value in var.allowed_values:
            etree.SubElement(v, 'allowedValue').text = value

        if var.allowed_value_range is not None and len(var.allowed_value_range) > 0:
          complete = True
          for name, value in var.allowed_value_range.items():
            if value is None:
              complete = False
          if complete:
            avl = etree.SubElement(s, 'allowedValueRange')
            for name, value in var.allowed_value_range.items():
              if value is not None:
                etree.SubElement(avl, name).text = str(value)

      return etree.tostring(root, encoding='utf-8', xml_declaration=True, pretty_print=True)

from twisted.python.util import OrderedDict


class ServiceControl(log.Loggable):

    def get_action_results(self, result, action, instance):
        """ check for out arguments
            if yes: check if there are related ones to StateVariables with
                    non A_ARG_TYPE_ prefix
                    if yes: check if there is a call plugin method for this action
                            if yes: update StateVariable values with call result
                            if no:  get StateVariable values and
                                    add them to result dict
        """
        self.debug('get_action_results %s', result)
        # print 'get_action_results', action, instance
        r = result
        notify = []
        for argument in action.get_out_arguments():
          # print 'get_state_variable_contents', argument.name
          if argument.name[0:11] != 'A_ARG_TYPE_':
            if action.get_callback() is not None:
              variable = self.variables[instance][argument.get_state_variable()]
              variable.update(r[argument.name])
              if variable.send_events == 'yes' and not variable.moderated:
                  notify.append(variable)
            else:
              variable = self.variables[instance][argument.get_state_variable()]
              r[argument.name] = variable.value
          self.service.propagate_notification(notify)

        self.info('action_results unsorted %s %s', action.name, r)
        if len(r) == 0:
            return r

        ordered_result = OrderedDict()
        for argument in action.get_out_arguments():
          ordered_result[argument.name] = r[argument.name]
        self.info('action_results sorted %s %s', action.name, ordered_result)
        return ordered_result

    def soap__generic(self, *args, **kwargs):
        """ generic UPnP service control method,
            which will be used if no soap_ACTIONNAME method
            in the server service control class can be found
        """
        try:
            action = self.actions[kwargs['soap_methodName']]
        except:
            return failure.Failure(errorCode(401))

        try:
            instance = int(kwargs['InstanceID'])
        except:
            instance = 0

        self.info("soap__generic %s %s %s", action, __name__, kwargs)
        del kwargs['soap_methodName']
        if(kwargs.has_key('X_UPnPClient') and
                kwargs['X_UPnPClient'] == 'XBox'):
            if(action.name == 'Browse' and
                    kwargs.has_key('ContainerID')):
                """ XXX: THIS IS SICK """
                kwargs['ObjectID'] = kwargs['ContainerID']
                del kwargs['ContainerID']

        in_arguments = action.get_in_arguments()
        for arg_name, arg in kwargs.iteritems():
            if arg_name.find('X_') == 0:
                continue
            l = [a for a in in_arguments if arg_name == a.get_name()]
            if len(l) > 0:
                in_arguments.remove(l[0])
            else:
                self.critical('argument %s not valid for action %s', arg_name, action.name)
                return failure.Failure(errorCode(402))
        if len(in_arguments) > 0:
            self.critical('argument %s missing for action %s',
                                [a.get_name() for a in in_arguments], action.name)
            return failure.Failure(errorCode(402))

        def callit(*args, **kwargs):
            #print 'callit args', args
            #print 'callit kwargs', kwargs
            result = {}
            callback = action.get_callback()
            if callback != None:
                return callback(**kwargs)
            return result

        def got_error(x):
            #print 'failure', x
            self.info('soap__generic error during call processing')
            return x

        # call plugin method for this action
        d = defer.maybeDeferred(callit, *args, **kwargs)
        d.addCallback(self.get_action_results, action, instance)
        d.addErrback(got_error)
        return d



"""Creates Github links from @user and #issue text.

Bascially a much simplified version of sphinxcontrib.issuetracker
with support for @user.
"""

import re
import sys

from docutils import nodes
from docutils.transforms import Transform

GITHUB_ISSUE_URL = "https://github.com/{0}/issues/{1}"
GITHUB_USER_URL = "https://github.com/{1}"

if sys.version_info[0] == 2:
  str = unicode


class GithubReferences(Transform):
  default_priority = 999

  def apply(self):
    config = self.document.settings.env.config
    issue_re = re.compile(config.github_issue_pattern)
    mention_re = re.compile(config.github_mention_pattern)

    self._replace_pattern(issue_re, GITHUB_ISSUE_URL)
    self._replace_pattern(mention_re, GITHUB_USER_URL)

  def _replace_pattern(self, pattern, url_format):
    project = self.document.settings.env.config.github_project

    for node in self.document.traverse(nodes.Text):
      parent = node.parent
      if isinstance(parent, (nodes.literal, nodes.FixedTextElement)):
        continue

      text = str(node)
      new_nodes = []
      last_ref_end = 0
      for match in pattern.finditer(text):
        head = text[last_ref_end:match.start()]
        if head:
          new_nodes.append(nodes.Text(head))

        last_ref_end = match.end()
        ref = url_format.format(project, match.group(1))
        link = nodes.reference(
          match.group(0),
          match.group(0),
          refuri=ref
        )
        new_nodes.append(link)

      if not new_nodes:
        continue

      tail = text[last_ref_end:]
      if tail:
        new_nodes.append(nodes.Text(tail))

      parent.replace(node, new_nodes)


def setup(app):
  app.add_config_value("github_project", None, "env")
  app.add_config_value("github_issue_pattern", r"#(\d+)", "env")
  app.add_config_value("github_mention_pattern", r"@(\w+)", "env")
  app.add_transform(GithubReferences)



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

import json
from twisted.web import resource, static
from twisted.internet import defer

from coherence import log


class JsonInterface(resource.Resource, log.Loggable):
    logCategory = 'json'

    def __init__(self, controlpoint):
        resource.Resource.__init__(self)
        log.Loggable.__init__(self)
        self.controlpoint = controlpoint
        self.controlpoint.coherence.add_web_resource('json', self)
        self.children = {}

    def render_GET(self, request):
        d = defer.maybeDeferred(self.do_the_render, request)
        return d

    def render_POST(self, request):
        d = defer.maybeDeferred(self.do_the_render, request)
        return d

    def getChildWithDefault(self, path, request):
        self.info('getChildWithDefault, %s, %s, %s %s %r', request.method, path, request.uri, request.client, request.args)
        #return self.do_the_render(request)
        d = defer.maybeDeferred(self.do_the_render, request)
        return d

    def do_the_render(self, request):
        self.warning('do_the_render, %s, %s, %s %r %s', request.method, request.path, request.uri, request.args, request.client)
        msg = "Houston, we've got a problem"
        path = request.path.split('/')
        path = path[2:]
        self.warning('path %r', path)
        if request.method in ('GET', 'POST'):
            request.postpath = None
            if request.method == 'GET':
                if path[0] == 'devices':
                    return self.list_devices(request)
                else:
                    device = self.controlpoint.get_device_with_id(path[0])
                    if device is not None:
                        service = device.get_service_by_type(path[1])
                        if service is not None:
                            action = service.get_action(path[2])
                            if action is not None:
                                return self.call_action(action, request)
                            else:
                                msg = "action %r on service type %r for device %r not found" % (path[2], path[1], path[0])
                        else:
                            msg = "service type %r for device %r not found" % (path[1], path[0])

                    else:
                        msg = "device with id %r not found" % path[0]

        request.setResponseCode(404, message=msg)
        return static.Data("<html><p>%s</p></html>" % msg, 'text/html')

    def list_devices(self, request):
        devices = []
        for device in self.controlpoint.get_devices():
            devices.append(device.as_dict())
        return static.Data(json.dumps(devices), 'application/json')

    def call_action(self, action, request):
        kwargs = {}
        for entry, value_list in request.args.items():
            kwargs[entry] = unicode(value_list[0])

        def to_json(result):
            self.warning("to_json")
            return static.Data(json.dumps(result), 'application/json')

        def fail(f):
            request.setResponseCode(404)
            return static.Data("<html><p>Houston, we've got a problem</p></html>", 'text/html')

        d = action.call(**kwargs)
        d.addCallback(to_json)
        d.addErrback(fail)
        return d



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2008, Frank Scholz <coherence@beebits.net>

"""
Test cases for L{dbus_service}
"""

import os

from twisted.trial import unittest
from twisted.internet import reactor
from twisted.internet.defer import Deferred

from coherence import __version__
from coherence.base import Coherence
from coherence.upnp.core import uuid
import coherence.extern.louie as louie

from tests import wrapped

try:
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)
    import dbus.service
except ImportError:
    dbus = None

BUS_NAME = 'org.Coherence'
OBJECT_PATH = '/org/Coherence'


class TestDBUS(unittest.TestCase):

    if not dbus:
        skip = "Python dbus-bindings not available."
    elif reactor.__class__.__name__ != 'Glib2Reactor':
        skip = ("This test needs a Glib2Reactor, please start trial "
                "with the '-r glib2' option.")

    def setUp(self):
        louie.reset()
        self.coherence = Coherence({'unittest': 'yes', 'logmode': 'error', 'use_dbus': 'yes', 'controlpoint': 'yes'})
        self.bus = dbus.SessionBus()
        self.coherence_service = self.bus.get_object(BUS_NAME, OBJECT_PATH)
        self.uuid = str(uuid.UUID())

    def tearDown(self):

        def cleaner(r):
            self.coherence.clear()
            return r

        dl = self.coherence.shutdown()
        dl.addBoth(cleaner)
        return dl

    def test_dbus_version(self):
        """ tests the version number request via dbus
        """

        d = Deferred()

        @wrapped(d)
        def handle_version_reply(version):
            self.assertEqual(version, __version__)
            d.callback(version)

        self.coherence_service.version(dbus_interface=BUS_NAME,
                                       reply_handler=handle_version_reply,
                                       error_handler=d.errback)
        return d

    def test_dbus_plugin_add_and_remove(self):
        """ tests creation and removal of a backend via dbus
        """

        d = Deferred()

        @wrapped(d)
        def add_it(uuid):
            self.coherence_service.add_plugin(
                'SimpleLight', {'name': 'dbus-test-light-%d' % os.getpid(), 'uuid': uuid},
                dbus_interface=BUS_NAME,
                reply_handler=handle_add_plugin_reply,
                error_handler=d.errback)

        @wrapped(d)
        def handle_add_plugin_reply(uuid):
            self.assertEqual(self.uuid, uuid)
            reactor.callLater(2, remove_it, uuid)

        @wrapped(d)
        def remove_it(uuid):
            self.coherence_service.remove_plugin(
                uuid,
                dbus_interface=BUS_NAME,
                reply_handler=handle_remove_plugin_reply,
                error_handler=d.errback)

        @wrapped(d)
        def handle_remove_plugin_reply(uuid):
            self.assertEqual(self.uuid, uuid)
            d.callback(uuid)

        add_it(self.uuid)
        return d



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2014, Hartmut Goebel <h.goebel@goebel-consult.de>

"""
Test cases for L{upnp.core.msearch}
"""

import time

from twisted.trial import unittest
from twisted.internet import protocol
from twisted.test import proto_helpers

from coherence.upnp.core import msearch, ssdp

SSDP_PORT = 1900
SSDP_ADDR = '239.255.255.250'

USN_1 = 'uuid:DDF542FB::urn:service:X_MS_MediaReceiverRegistrar:1'
MSEARCH_RESPONSE_1 = (
    'HTTP/1.1 200 OK',
    'Cache-Control: max-age=1802',
    'Date: Mon, 02 Jun 2014 18:04:59 GMT',
    'Ext: ',
    'Location: http://192.168.1.4:9000/plugins/UPnP/MediaServer.xml',
    'Server: Linux/armv5-linux UPnP/1.0 DLNADOC/1.50 MediaServer/7.3/735',
    'ST: urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:1',
    'USN: ' + USN_1,
    )


class TestMSearch(unittest.TestCase):

    def setUp(self):
        self.tr = proto_helpers.FakeDatagramTransport()
        self.ssdp_server = ssdp.SSDPServer(test=True)
        self.ssdp_server.makeConnection(self.tr)
        self.proto = msearch.MSearch(self.ssdp_server, test=True)
        self.proto.makeConnection(self.tr)

    def test_msearch_request(self):
        self.assertEqual(self.ssdp_server.known, {})
        data = '\r\n'.join(MSEARCH_RESPONSE_1) + '\r\n\r\n'
        self.proto.datagramReceived(data, ('10.20.30.40', 1234))
        self.assertTrue(self.ssdp_server.isKnown(USN_1))
        self.assertFalse(self.ssdp_server.isKnown(USN_1*2))
        service = self.ssdp_server.known[USN_1]
        del service['last-seen']
        self.assertEqual(service, {
            'HOST': '10.20.30.40',
            'ST': 'urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:1',
            'LOCATION': 'http://192.168.1.4:9000/plugins/UPnP/MediaServer.xml',
            'USN': USN_1,
            'CACHE-CONTROL': 'max-age=1802',
            'SERVER': 'Linux/armv5-linux UPnP/1.0 DLNADOC/1.50 MediaServer/7.3/735',
            'MANIFESTATION': 'remote',
            'SILENT': False,
            'EXT': '',
            })

    def test_msearch_request_does_not_send_reply(self):
        data = '\r\n'.join(MSEARCH_RESPONSE_1) + '\r\n\r\n'
        self.proto.datagramReceived(data, ('10.20.30.40', 1234))
        self.assertEqual(self.tr.written, [])

    def test_msearch_request_updates_timestamp(self):
        data = '\r\n'.join(MSEARCH_RESPONSE_1) + '\r\n\r\n'
        self.proto.datagramReceived(data, ('10.20.30.40', 1234))
        service1 = self.ssdp_server.known[USN_1]
        last_seen1 = service1['last-seen']
        time.sleep(0.5)
        self.proto.datagramReceived(data, ('10.20.30.40', 1234))
        service2 = self.ssdp_server.known[USN_1]
        last_seen2 = service1['last-seen']
        self.assertIs(service1, service2)
        self.assertLess(last_seen1, last_seen2+0.5)

    def test_discover(self):
        self.assertEqual(self.tr.written, [])
        self.proto.discover()
        expected = [(l + '\r\n') for l in [
            'M-SEARCH * HTTP/1.1',
            'HOST: 239.255.255.250:1900',
            'MAN: "ssdp:discover"',
            'MX: 5',
            'ST: ssdp:all',
            ''
            ]]
        # :fixme: What is the reason the notification is send twice?
        self.assertEqual(len(self.tr.written), 1)
        data, (host, port) = self.tr.written[0]
        self.assertEqual((host, port), (SSDP_ADDR, SSDP_PORT))
        recieved = data.splitlines(True)
        self.assertEqual(sorted(recieved), sorted(expected))



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2007 - Frank Scholz <coherence@beebits.net>

""" SOAP-lite

    some simple functions to implement the SOAP msgs
    needed by UPnP with ElementTree

    inspired by ElementSOAP.py
"""
import logging
from coherence.upnp.core.xml_constants import ELEMENT_TYPE
from twisted.python.util import OrderedDict

from lxml import etree

NS_SOAP_ENV = "http://schemas.xmlsoap.org/soap/envelope/"
NS_SOAP_ENC = "http://schemas.xmlsoap.org/soap/encoding/"
NS_XSD = "http://www.w3.org/1999/XMLSchema"
NS_UPNP_ORG_CONTROL_1_0 = 'urn:schemas-upnp-org:control-1-0'

TYPE_MAP = {str: 'string',
            unicode: 'string',
            int: 'int',
            float: 'float',
            bool: 'boolean'}

UPNPERRORS = {401: 'Invalid Action',
              402: 'Invalid Args',
              501: 'Action Failed',
              600: 'Argument Value Invalid',
              601: 'Argument Value Out of Range',
              602: 'Optional Action Not Implemented',
              603: 'Out Of Memory',
              604: 'Human Intervention Required',
              605: 'String Argument Too Long',
              606: 'Action Not Authorized',
              607: 'Signature Failure',
              608: 'Signature Missing',
              609: 'Not Encrypted',
              610: 'Invalid Sequence',
              611: 'Invalid Control URL',
              612: 'No Such Session', }

logger = logging.getLogger('soap_lite')


def build_soap_error(status,
                     description='without words',
                     pretty_print=True):
  """ builds an UPnP SOAP error msg
  """
  root = etree.Element(etree.QName(NS_SOAP_ENV, 'Fault'))
  etree.SubElement(root, 'faultcode').text = 's:Client'
  etree.SubElement(root, 'faultstring').text = 'UPnPError'
  e = etree.SubElement(root, 'detail')
  e = etree.SubElement(e, etree.QName(NS_UPNP_ORG_CONTROL_1_0, 'UPnPError'), nsmap={None: NS_UPNP_ORG_CONTROL_1_0})
  etree.SubElement(e, 'errorCode').text = str(status)
  etree.SubElement(e, 'errorDescription').text = UPNPERRORS.get(status, description)

  return build_soap_call(None, root, pretty_print=pretty_print)


def build_soap_call(method, arguments, ns=None,
                    is_response=False,
                    pretty_print=True):
  """ create a shell for a SOAP request or response element
      - set method to none to omitt the method element and
        add the arguments directly to the body (for an error msg)
      - arguments can be a dict or an etree.Element
  """
  envelope = etree.Element(etree.QName(NS_SOAP_ENV, 'Envelope'),
                           attrib={etree.QName(NS_SOAP_ENV, 'encodingStyle'): NS_SOAP_ENC},
                           nsmap={'s': NS_SOAP_ENV})
  body = etree.SubElement(envelope, etree.QName(NS_SOAP_ENV, 'Body'))

  if method:
    if is_response is True:
      method += "Response"

    if ns:
      tag = etree.QName(ns, method)
      nsmap = {'u': ns}
    else:
      tag = method
      nsmap = None

    re = etree.SubElement(body, tag, nsmap=nsmap)
  else:
    re = body

  # append the arguments
  if isinstance(arguments, (dict, OrderedDict)):
    for arg_name, arg_val in arguments.iteritems():
      if type(arg_val) in TYPE_MAP:
        arg_type = TYPE_MAP[type(arg_val)]
        if arg_type == 'int' or arg_type == 'float':
          arg_val = str(arg_val)
        if arg_type == 'boolean':
          arg_val = '1' if arg_val else '0'
        e = etree.SubElement(re, arg_name)
        e.text = arg_val
      # elif isinstance(arg_val, ELEMENT_TYPE):
      #   e.append(arg_val)
  elif isinstance(arguments, ELEMENT_TYPE):
    re.append(arguments)

  xml = etree.tostring(envelope, encoding='utf-8', xml_declaration=True, pretty_print=pretty_print)
  logger.debug("xml dump:\n%s", xml)
  return xml



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2007, Frank Scholz <coherence@beebits.net>

"""

Covers by Amazon

methods to retrieve covers/album art via the
Amazon E-Commerce WebService v4
http://docs.amazonwebservices.com/AWSECommerceService/2007-04-04/DG/

The licence agreement says something about only
one request per second, so we need to serialize
and delay the calls a bit.

The AWSAccessKeyId supplied is _ONLY_ for the use
in conjunction with Coherence, http://coherence.beebits.net

If you use this library in your own software please
apply for your own key @ http://www.amazon.com/webservices
and follow the rules of their license.

Especially you must add the following disclaimer in a place
that is reasonably viewable by the user of your application:

 PLEASE KEEP IN MIND THAT SOME OF THE CONTENT THAT WE
 MAKE AVAILABLE TO YOU THROUGH THIS APPLICATION COMES
 FROM AMAZON WEB SERVICES. ALL SUCH CONTENT IS PROVIDED
 TO YOU "AS IS." THIS CONTENT AND YOUR USE OF IT
 ARE SUBJECT TO CHANGE AND/OR REMOVAL AT ANY TIME.

Furthermore if you save any of the cover images you
have to take care that they are stored no longer than
a maximum of one month and requested then from Amazon
again.

"""
from lxml import etree

import os
import urllib
import StringIO

from twisted.internet import reactor
from twisted.internet import defer
from twisted.web import client

aws_server = {'de': 'de',
              'jp': 'jp',
              'ca': 'ca',
              'uk': 'co.uk',
              'fr': 'fr'}

aws_artist_query = '&Operation=ItemSearch' \
                   '&SearchIndex=Music'

aws_asin_query = '&Operation=ItemLookup'

aws_response_group = '&ResponseGroup=Images'

aws_ns = 'http://webservices.amazon.com/AWSECommerceService/2005-10-05'

aws_image_size = {'large': 'LargeImage',
                   'medium': 'MediumImage',
                   'small': 'SmallImage'}


class WorkQueue(object):

    _instance_ = None  # Singleton

    def __new__(cls, *args, **kwargs):
        obj = getattr(cls, '_instance_', None)
        if obj is not None:
            return obj
        else:
            obj = super(WorkQueue, cls).__new__(cls, *args, **kwargs)
            cls._instance_ = obj
            obj.max_workers = kwargs.get('max_workers', 1)
            obj.queue = []
            obj.workers = []
            return obj

    def __init__(self, method, *args, **kwargs):
        self.queue.append((method, args, kwargs))
        self.queue_run()

    def queue_run(self):
        if len(self.queue) == 0:
            return
        if len(self.workers) >= self.max_workers:
            #print "WorkQueue - all workers busy"
            return
        work = self.queue.pop()
        d = defer.maybeDeferred(work[0], *work[1], **work[2])
        self.workers.append(d)
        d.addCallback(self.remove_from_workers, d)
        d.addErrback(self.remove_from_workers, d)

    def remove_from_workers(self, result, d):
        self.workers.remove(d)
        reactor.callLater(1, self.queue_run)  # a very,very weak attempt


class CoverGetter(object):

    """
    retrieve a cover image for a given ASIN,
                               a TITLE or
                               an ARTIST/TITLE combo

    parameters are:

        filename: where to save a received image
                  if NONE the image will be passed to the callback
        callback: a method to call with the filename
                  or the image as a parameter
                  after the image request and save was successful
                  can be:
                  - only a callable
                  - a tuple with a callable,
                      - optional an argument or a tuple of arguments
                      - optional a dict with keyword arguments
        not_found_callback: a method to call when the search at Amazon failed
                  can be:
                  - only a callable
                  - a tuple with a callable,
                      - optional an argument or a tuple of arguments
                      - optional a dict with keyword arguments
        locale:   which Amazon Webservice Server to use, defaults to .com
        image_size: request the cover as large|medium|small image
                    resolution seems to be in pixels for
                    large: 500x500, medium: 160x160 and small: 75x75
        asin: the Amazon Store Identification Number
        artist: the artists name
        title: the album title

    if the filename extension and the received image extension differ,
    the image is converted with PIL to the desired format
    http://www.pythonware.com/products/pil/index.htm

    """

    def __init__(self, filename, aws_key, callback=None, not_found_callback=None,
                       locale=None,
                       image_size='large',
                       title=None, artist=None, asin=None):
        self.aws_base_query = '/onca/xml?Service=AWSECommerceService' \
                              '&AWSAccessKeyId=%s' % aws_key

        self.filename = filename
        self.callback = callback
        self._errcall = not_found_callback
        self.server = 'http://ecs.amazonaws.%s' % aws_server.get(locale, 'com')
        self.image_size = image_size

        def sanitize(s):
            if s is not None:
                s = unicode(s.lower())
                s = s.replace(unicode(u'ä'), unicode('ae'))
                s = s.replace(unicode(u'ö'), unicode('oe'))
                s = s.replace(unicode(u'ü'), unicode('ue'))
                s = s.replace(unicode(u'ß'), unicode('ss'))
                if isinstance(s, unicode):
                    s = s.encode('ascii', 'ignore')
                else:
                    s = s.decode('utf-8').encode('ascii', 'ignore')
            return s

        if asin != None:
            query = aws_asin_query + '&ItemId=%s' % urllib.quote(asin)
        elif (artist is not None or title is not None):
            query = aws_artist_query
            if artist is not None:
                artist = sanitize(artist)
                query = '&'.join((query, 'Artist=%s' % urllib.quote(artist)))
            if title is not None:
                title = sanitize(title)
                query = '&'.join((query, 'Title=%s' % urllib.quote(title)))
        else:
            raise KeyError("Please supply either asin, title or artist and title arguments")
        url = self.server + self.aws_base_query + aws_response_group + query
        WorkQueue(self.send_request, url)

    def send_request(self, url, *args, **kwargs):
        #print "send_request", url
        d = client.getPage(url)
        d.addCallback(self.got_response)
        d.addErrback(self.got_error, url)
        return d

    def got_image(self, result, convert_from='', convert_to=''):
        #print "got_image"
        if(len(convert_from) and len(convert_to)):
            #print "got_image %d, convert to %s" % (len(result), convert_to)
            try:
                import Image

                im = Image.open(StringIO.StringIO(result))
                name, file_ext = os.path.splitext(self.filename)
                self.filename = name + convert_to

                im.save(self.filename)
            except ImportError:
                print "we need the Python Imaging Library to do image conversion"

        if self.filename == None:
            data = result
        else:
            data = self.filename

        if self.callback is not None:
            #print "got_image", self.callback
            if isinstance(self.callback, tuple):
                if len(self.callback) == 3:
                    c, a, kw = self.callback
                    if not isinstance(a, tuple):
                        a = (a, )
                    a = (data, ) + a
                    c(*a, **kw)
                if len(self.callback) == 2:
                    c, a = self.callback
                    if isinstance(a, dict):
                        c(data, **a)
                    else:
                        if not isinstance(a, tuple):
                            a = (a, )
                        a = (data, ) + a
                        c(*a)
                if len(self.callback) == 1:
                    c = self.callback
                    c(data)
            else:
                self.callback(data)

    def got_response(self, result):
        convert_from = convert_to = ''
        result = etree.fromstring(result)
        image_tag = result.find('.//{%s}%s' % (aws_ns, aws_image_size.get(self.image_size, 'large')))
        if image_tag is not None:
            image_url = image_tag.findtext('{%s}URL' % aws_ns)
            if self.filename is None:
                d = client.getPage(image_url)
            else:
                _, file_ext = os.path.splitext(self.filename)
                if file_ext == '':
                    _, image_ext = os.path.splitext(image_url)
                    if image_ext != '':
                        self.filename = ''.join((self.filename, image_ext))
                else:
                    _, image_ext = os.path.splitext(image_url)
                    if image_ext != '' and file_ext != image_ext:
                        #print "hmm, we need a conversion..."
                        convert_from = image_ext
                        convert_to = file_ext
                if len(convert_to):
                    d = client.getPage(image_url)
                else:
                    d = client.downloadPage(image_url, self.filename)
            d.addCallback(self.got_image, convert_from=convert_from, convert_to=convert_to)
            d.addErrback(self.got_error, image_url)
        else:
            if self._errcall is not None:
                if isinstance(self._errcall, tuple):
                    if len(self._errcall) == 3:
                        c, a, kw = self._errcall
                        if not isinstance(a, tuple):
                            a = (a, )
                        c(*a, **kw)
                    if len(self._errcall) == 2:
                        c, a = self._errcall
                        if isinstance(a, dict):
                            c(**a)
                        else:
                            if not isinstance(a, tuple):
                                a = (a, )
                            c(*a)
                    if len(self._errcall) == 1:
                        c = self._errcall
                        c()
                else:
                    self._errcall()

    def got_error(self, failure, url):
        print "got_error", failure, url

if __name__ == '__main__':

    from twisted.python import usage

    class Options(usage.Options):
        optParameters = [['artist', 'a', '', 'artist name'],
                         ['title', 't', '', 'title'],
                         ['asin', 's', '', 'ASIN'],
                         ['filename', 'f', 'cover.jpg', 'filename'],
                    ]

    options = Options()
    try:
        options.parseOptions()
    except usage.UsageError, errortext:
        import sys
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)

    def got_it(filename, *args, **kwargs):
        print "Mylady, it is an image and its name is", filename, args, kwargs

    aws_key = '1XHSE4FQJ0RK0X3S9WR2'
    print options['asin'], options['artist'], options['title']
    if len(options['asin']):
        reactor.callWhenRunning(CoverGetter, options['filename'], aws_key, callback=got_it, asin=options['asin'])
    elif len(options['artist']) and len(options['title']):
        reactor.callWhenRunning(CoverGetter, options['filename'], aws_key, callback=got_it, artist=options['artist'], title=options['title'])

    reactor.run()



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2014 Hartmut Goebel <h.goebel@goebel-consult.de>

"""
Test cases for L{upnp.backends.fs_storage}
"""

from twisted.trial import unittest
from twisted.python.filepath import FilePath

from coherence.backends import fs_storage

import coherence.log
coherence.log.init()

class TestFSStorageAssumptions(unittest.TestCase):

    def setUp(self):
        self.tmp_content = FilePath(self.mktemp())
        self.tmp_content.makedirs()
        self.storage = fs_storage.FSStore(None, name='my media',
                                          content=self.tmp_content.path,
                                          urlbase='http://fsstore-host/xyz',
                                          enable_inotify=False)

    def tearDown(self):
        self.tmp_content.remove()
        pass

    def test_ContentLen(self):
        self.assertEqual(len(self.storage.content), 1)
        self.assertEqual(len(self.storage.store), 1)
        self.assertEqual(self.storage.len(), 1)

    def test_Root(self):
        root = self.storage.get_by_id('1000')
        self.assertIs(root.parent, None)
        self.assertRaises(AttributeError, getattr, root, 'path')
        # A single path passed, so content is a "directory" named by
        # it's basename
        self.assertEqual(root.mimetype, 'directory')
        self.assertEqual(root.get_name(), 'temp')


class TestFSStorageWithMultiContentAssumptions(unittest.TestCase):

    def setUp(self):
        f = self.tmp_content = FilePath(self.mktemp())
        audio = f.child('audio') ; audio.makedirs()
        video = f.child('video') ; video.makedirs()
        self.storage = fs_storage.FSStore(None, name='my media',
                                          content=[audio.path, video.path],
                                          urlbase='http://fsstore-host/xyz',
                                          enable_inotify=False)

    def tearDown(self):
        self.tmp_content.remove()
        pass

    def test_ContentLen(self):
        self.assertEqual(len(self.storage.content), 2)
        self.assertEqual(len(self.storage.store), 3)
        self.assertEqual(self.storage.len(), 3)

    def test_Root(self):
        root = self.storage.get_by_id('1000')
        self.assertIs(root.parent, None)
        self.assertRaises(AttributeError, getattr, root, 'path')
        # A several paths passed, so content is a "root" named "media"
        self.assertEqual(root.mimetype, 'root')
        self.assertEqual(root.get_name(), 'media')

# todo: test get_xml()
'''
self.storage.get_by_id("1000").get_xml()
<container xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" childCount="2" id="1000" parentID="-1" restricted="0"><dc:title>media</dc:title><upnp:class>object.container</upnp:class><dc:date>2003-07-23T01:18:00+02:00</dc:date></container>
'''

class TestFSStorage(unittest.TestCase):

    def setUp(self):
        self.tmp_content = FilePath(self.mktemp())
        f = self.tmp_content.child('my content')
        audio = f.child('audio') ; audio.makedirs()
        video = f.child('video') ; video.makedirs()
        images = f.child('images') ; images.makedirs()
        album = audio.child('album-1')
        album.makedirs()
        album.child('track-1.mp3').touch()
        album.child('track-2.mp3').touch()
        album = audio.child('album-2')
        album.makedirs()
        album.child('track-1.ogg').touch()
        album.child('track-2.ogg').touch()
        self.storage = fs_storage.FSStore(None, name='my media',
                                          content=self.tmp_content.path,
                                          urlbase='http://fsstore-host/xyz',
                                          enable_inotify=False)

    def tearDown(self):
        self.tmp_content.remove()

    def test_ContentLen(self):
        self.assertEqual(len(self.storage.content), 1)
        # 11 items, since we have "<tempdir>/my content/..."
        self.assertEqual(len(self.storage.store), 11)
        self.assertEqual(self.storage.len(), 11)

    def test_Content(self):
        root = self.storage.get_by_id('1000')
        content = self.storage.get_by_id('1001')
        self.assertEqual(content.mimetype, 'directory')
        self.assertEqual(content.get_name(), 'my content')
        self.assertIs(root.get_children(0, 0)[0], content)
        self.assertEqual(self.storage.get_by_id('1002').get_name(),
                         'audio')
        self.assertEqual(self.storage.get_by_id('1005').get_name(),
                         'album-1')



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2005, Tim Potter <tpot@samba.org>
# Copyright 2006 John-Mark Gurney <gurney_j@resnet.uoregon.edu>
# Copyright 2006, Frank Scholz <coherence@beebits.net>

# Content Directory service

from twisted.python import failure
from twisted.web import resource
from twisted.internet import defer

from coherence.upnp.core.soap_service import UPnPPublisher
from coherence.upnp.core.soap_service import errorCode
from coherence.upnp.core.DIDLLite import DIDLElement

from coherence.upnp.core import service


class ContentDirectoryControl(service.ServiceControl, UPnPPublisher):

    def __init__(self, server):
        service.ServiceControl.__init__(self)
        UPnPPublisher.__init__(self)
        self.service = server
        self.variables = server.get_variables()
        self.actions = server.get_actions()


class ContentDirectoryServer(service.ServiceServer, resource.Resource):
    logCategory = 'content_directory_server'

    def __init__(self, device, backend=None, transcoding=False):
        self.device = device
        self.transcoding = transcoding
        if backend is None:
            backend = self.device.backend
        resource.Resource.__init__(self)
        service.ServiceServer.__init__(self, 'ContentDirectory', self.device.version, backend)

        self.control = ContentDirectoryControl(self)
        self.putChild('scpd.xml', service.scpdXML(self, self.control))
        self.putChild('control', self.control)

        self.set_variable(0, 'SystemUpdateID', 0)
        self.set_variable(0, 'ContainerUpdateIDs', '')

    def listchilds(self, uri):
        cl = ''
        for c in self.children:
            cl += '<li><a href=%s/%s>%s</a></li>' % (uri, c, c)
        return cl

    def render(self, request):
        return '<html><p>root of the ContentDirectory</p><p><ul>%s</ul></p></html>' % self.listchilds(request.uri)

    def upnp_Search(self, *args, **kwargs):
        ContainerID = kwargs['ContainerID']
        Filter = kwargs['Filter']
        StartingIndex = int(kwargs['StartingIndex'])
        RequestedCount = int(kwargs['RequestedCount'])
        SortCriteria = kwargs['SortCriteria']
        SearchCriteria = kwargs['SearchCriteria']

        total = 0
        root_id = 0
        item = None
        items = []

        parent_container = str(ContainerID)

        didl = DIDLElement(upnp_client=kwargs.get('X_UPnPClient', ''),
                           parent_container=parent_container,
                           transcoding=self.transcoding)

        def build_response(tm):
            r = {'Result': didl.toString(), 'TotalMatches': tm,
                 'NumberReturned': didl.numItems()}

            if hasattr(item, 'update_id'):
                r['UpdateID'] = item.update_id
            elif hasattr(self.backend, 'update_id'):
                r['UpdateID'] = self.backend.update_id  # FIXME
            else:
                r['UpdateID'] = 0

            return r

        def got_error(r):
            return r

        def process_result(result, total=None, found_item=None):
            if result == None:
                result = []

            l = []

            def process_items(result, tm):
                if result is None:
                    result = []
                for i in result:
                    if i[0]:
                        didl.addItem(i[1])

                return build_response(tm)

            for i in result:
                d = defer.maybeDeferred(i.get_item)
                l.append(d)

            if found_item is not None:
                def got_child_count(count):
                    dl = defer.DeferredList(l)
                    dl.addCallback(process_items, count)
                    return dl

                d = defer.maybeDeferred(found_item.get_child_count)
                d.addCallback(got_child_count)

                return d
            elif total is None:
                total = item.get_child_count()

            dl = defer.DeferredList(l)
            dl.addCallback(process_items, total)
            return dl

        def proceed(result):
            if(kwargs.get('X_UPnPClient', '') == 'XBox' and
               hasattr(result, 'get_artist_all_tracks')):
                d = defer.maybeDeferred(result.get_artist_all_tracks, StartingIndex, StartingIndex + RequestedCount)
            else:
                d = defer.maybeDeferred(result.get_children, StartingIndex, StartingIndex + RequestedCount)
            d.addCallback(process_result, found_item=result)
            d.addErrback(got_error)
            return d

        try:
            root_id = ContainerID
        except:
            pass

        wmc_mapping = getattr(self.backend, "wmc_mapping", None)
        if kwargs.get('X_UPnPClient', '') == 'XBox':
            if(wmc_mapping is not None and
               wmc_mapping.has_key(ContainerID)):
                """ fake a Windows Media Connect Server
                """
                root_id = wmc_mapping[ContainerID]
                if callable(root_id):
                    item = root_id()
                    if item is not None:
                        if isinstance(item, list):
                            total = len(item)
                            if int(RequestedCount) == 0:
                                items = item[StartingIndex:]
                            else:
                                items = item[StartingIndex:StartingIndex + RequestedCount]
                            return process_result(items, total=total)
                        else:
                            if isinstance(item, defer.Deferred):
                                item.addCallback(proceed)
                                return item
                            else:
                                return proceed(item)

                item = self.backend.get_by_id(root_id)
                if item is None:
                    return process_result([], total=0)

                if isinstance(item, defer.Deferred):
                    item.addCallback(proceed)
                    return item
                else:
                    return proceed(item)

        item = self.backend.get_by_id(root_id)
        if item is None:
            return failure.Failure(errorCode(701))

        if isinstance(item, defer.Deferred):
            item.addCallback(proceed)
            return item
        else:
            return proceed(item)

    def upnp_Browse(self, *args, **kwargs):
        try:
            ObjectID = kwargs['ObjectID']
        except:
            self.debug("hmm, a Browse action and no ObjectID argument? An XBox maybe?")
            try:
                ObjectID = kwargs['ContainerID']
            except:
                ObjectID = 0
        BrowseFlag = kwargs['BrowseFlag']
        Filter = kwargs['Filter']
        StartingIndex = int(kwargs['StartingIndex'])
        RequestedCount = int(kwargs['RequestedCount'])
        SortCriteria = kwargs['SortCriteria']
        parent_container = None
        requested_id = None

        item = None
        total = 0
        items = []


        if BrowseFlag == 'BrowseDirectChildren':
            parent_container = str(ObjectID)
        else:
            requested_id = str(ObjectID)

        self.info("upnp_Browse request %r %r %r %r", ObjectID, BrowseFlag, StartingIndex, RequestedCount)

        didl = DIDLElement(upnp_client=kwargs.get('X_UPnPClient', ''),
                           requested_id=requested_id,
                           parent_container=parent_container,
                           transcoding=self.transcoding)

        def got_error(r):
            return r

        def process_result(result, total=None, found_item=None):
            if result == None:
                result = []
            if BrowseFlag == 'BrowseDirectChildren':
                l = []

                def process_items(result, tm):
                    if result == None:
                        result = []
                    for i in result:
                        if i[0] == True:
                            didl.addItem(i[1])

                    return build_response(tm)

                for i in result:
                    d = defer.maybeDeferred(i.get_item)
                    l.append(d)

                if found_item != None:
                    def got_child_count(count):
                        dl = defer.DeferredList(l)
                        dl.addCallback(process_items, count)
                        return dl

                    d = defer.maybeDeferred(found_item.get_child_count)
                    d.addCallback(got_child_count)

                    return d
                elif total == None:
                    total = item.get_child_count()

                dl = defer.DeferredList(l)
                dl.addCallback(process_items, total)
                return dl
            else:
                didl.addItem(result)
                total = 1

            return build_response(total)

        def build_response(tm):
          r = {'Result': didl.toString(),
               'TotalMatches': tm,
               'NumberReturned': didl.numItems()}

          if hasattr(item, 'update_id'):
            r['UpdateID'] = item.update_id
          elif hasattr(self.backend, 'update_id'):
            r['UpdateID'] = self.backend.update_id  # FIXME
          else:
            r['UpdateID'] = 0

          return r

        def proceed(result):
            if BrowseFlag == 'BrowseDirectChildren':
                d = defer.maybeDeferred(result.get_children, StartingIndex, StartingIndex + RequestedCount)
            else:
                d = defer.maybeDeferred(result.get_item)

            d.addCallback(process_result, found_item=result)
            d.addErrback(got_error)
            return d

        root_id = ObjectID

        wmc_mapping = getattr(self.backend, "wmc_mapping", None)
        if kwargs.get('X_UPnPClient', '') == 'XBox' and \
                wmc_mapping is not None and wmc_mapping.has_key(ObjectID):
            """ fake a Windows Media Connect Server
            """
            root_id = wmc_mapping[ObjectID]
            if callable(root_id):
                item = root_id()
                if item is not None:
                    if isinstance(item, list):
                        total = len(item)
                        if int(RequestedCount) == 0:
                            items = item[StartingIndex:]
                        else:
                            items = item[StartingIndex:StartingIndex + RequestedCount]
                        return process_result(items, total=total)
                    else:
                        if isinstance(item, defer.Deferred):
                            item.addCallback(proceed)
                            return item
                        else:
                            return proceed(item)

            item = self.backend.get_by_id(root_id)
            if item is None:
                return process_result([], total=0)

            if isinstance(item, defer.Deferred):
                item.addCallback(proceed)
                return item
            else:
                return proceed(item)

        item = self.backend.get_by_id(root_id)
        if item is None:
            return failure.Failure(errorCode(701))

        if isinstance(item, defer.Deferred):
            item.addCallback(proceed)
            return item
        else:
            return proceed(item)



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2007, Frank Scholz <coherence@beebits.net>
# Copyright 2008, Jean-Michel Sizun <jm.sizun@free.fr>

from twisted.internet import defer

from coherence.upnp.core import utils
from coherence.upnp.core.utils import ReverseProxyUriResource

from coherence.upnp.core.DIDLLite import classChooser, Container, Resource, DIDLElement

from coherence.backend import BackendStore
from coherence.backend import BackendItem

from urlparse import urlsplit

from coherence.extern.galleryremote import Gallery


class ProxyGallery2Image(ReverseProxyUriResource):

    def __init__(self, uri):
        ReverseProxyUriResource.__init__(self, uri)

    def render(self, request):
        del request.received_headers['referer']
        return ReverseProxyUriResource.render(self, request)


class Gallery2Item(BackendItem):
    logCategory = 'gallery2_item'

    def __init__(self, id, obj, parent, mimetype, urlbase, UPnPClass, update=False):
        BackendItem.__init__(self)
        self.id = id

        self.name = obj.get('title')  # .encode('utf-8')
        if self.name == None:
            self.name = obj.get('name')
        if self.name == None:
            self.name = id

        self.mimetype = mimetype

        self.gallery2_id = obj.get('gallery2_id')

        self.parent = parent
        if parent:
            parent.add_child(self, update=update)

        if parent == None:
            parent_id = -1
        else:
            parent_id = parent.get_id()

        self.item = UPnPClass(id, parent_id, self.name)
        if isinstance(self.item, Container):
            self.item.childCount = 0
        self.child_count = 0
        self.children = None

        if(len(urlbase) and urlbase[-1] != '/'):
            urlbase += '/'

        if parent == None:
            self.gallery2_url = None
            self.url = urlbase + str(self.id)
        elif self.mimetype == 'directory':
            #self.gallery2_url = parent.store.get_url_for_album(self.gallery2_id)
            self.url = urlbase + str(self.id)
        else:
            self.gallery2_url = parent.store.get_url_for_image(self.gallery2_id)
            self.url = urlbase + str(self.id)
            self.location = ProxyGallery2Image(self.gallery2_url)

        if self.mimetype == 'directory':
            self.update_id = 0
        else:
            res = Resource(self.gallery2_url, 'http-get:*:%s:*' % self.mimetype)
            res.size = None
            self.item.res.append(res)

    def remove(self):
        if self.parent:
            self.parent.remove_child(self)
        del self.item

    def add_child(self, child, update=False):
        if self.children == None:
            self.children = []
        self.children.append(child)
        self.child_count += 1
        if isinstance(self.item, Container):
            self.item.childCount += 1
        if update == True:
            self.update_id += 1

    def remove_child(self, child):
        #self.info("remove_from %d (%s) child %d (%s)" % (self.id, self.get_name(), child.id, child.get_name()))
        if child in self.children:
            self.child_count -= 1
            if isinstance(self.item, Container):
                self.item.childCount -= 1
            self.children.remove(child)
            self.update_id += 1

    def get_children(self, start=0, request_count=0):
        def process_items(result=None):
            if self.children == None:
                return  []
            if request_count == 0:
                return self.children[start:]
            else:
                return self.children[start:request_count]

        if (self.children == None):
            d = self.store.retrieveItemsForAlbum(self.gallery2_id, self)
            d.addCallback(process_items)
            return d
        else:
            return process_items()

    def get_child_count(self):
        return self.child_count

    def get_id(self):
        return self.id

    def get_update_id(self):
        if hasattr(self, 'update_id'):
            return self.update_id
        else:
            return None

    def get_path(self):
        return self.url

    def get_name(self):
        return self.name

    def get_parent(self):
        return self.parent

    def get_item(self):
        return self.item

    def get_xml(self):
        return self.item.toString()

    def __repr__(self):
        return 'id: ' + str(self.id)


class Gallery2Store(BackendStore):

    logCategory = 'gallery2_store'

    implements = ['MediaServer']

    description = ('Gallery2', 'exposes the photos from a Gallery2 photo repository.', None)

    options = [{'option': 'name', 'text': 'Server Name:', 'type': 'string', 'default': 'my media', 'help': 'the name under this MediaServer shall show up with on other UPnP clients'},
       {'option': 'version', 'text': 'UPnP Version:', 'type': 'int', 'default': 2, 'enum': (2, 1), 'help': 'the highest UPnP version this MediaServer shall support', 'level': 'advance'},
       {'option': 'uuid', 'text': 'UUID Identifier:', 'type': 'string', 'help': 'the unique (UPnP) identifier for this MediaServer, usually automatically set', 'level': 'advance'},
       {'option': 'server_url', 'text': 'Server URL:', 'type': 'string'},
       {'option': 'username', 'text': 'User ID:', 'type': 'string', 'group': 'User Account'},
       {'option': 'password', 'text': 'Password:', 'type': 'string', 'group': 'User Account'},
    ]

    def __init__(self, server, **kwargs):
        BackendStore.__init__(self, server, **kwargs)

        self.next_id = 1000
        self.config = kwargs
        self.name = kwargs.get('name', 'gallery2Store')

        self.wmc_mapping = {'16': 1000}

        self.update_id = 0
        self.store = {}

        self.gallery2_server_url = self.config.get('server_url', 'http://localhost/gallery2')
        self.gallery2_username = self.config.get('username', None)
        self.gallery2_password = self.config.get('password', None)

        self.store[1000] = Gallery2Item(1000, {'title': 'Gallery2', 'gallery2_id': '0', 'mimetype': 'directory'}, None,
                                                        'directory', self.urlbase, Container, update=True)
        self.store[1000].store = self

        self.gallery2_remote = Gallery(self.gallery2_server_url, 2)
        if not None in [self.gallery2_username, self.gallery2_password]:
            d = self.gallery2_remote.login(self.gallery2_username, self.gallery2_password)
            d.addCallback(lambda x: self.retrieveAlbums('0', self.store[1000]))
            d.addCallback(self.init_completed)
        else:
            d = self.retrieveAlbums('0', self.store[1000])
            d.addCallback(self.init_completed)

    def __repr__(self):
        return self.__class__.__name__

    def append(self, obj, parent):
        if isinstance(obj, basestring):
            mimetype = 'directory'
        else:
            mimetype = obj['mimetype']

        UPnPClass = classChooser(mimetype)
        id = self.getnextID()
        update = False
        #if hasattr(self, 'update_id'):
        #    update = True

        item = Gallery2Item(id, obj, parent, mimetype, self.urlbase,
                                        UPnPClass, update=update)
        self.store[id] = item
        self.store[id].store = self
        if hasattr(self, 'update_id'):
            self.update_id += 1
            if self.server:
                self.server.content_directory_server.set_variable(0, 'SystemUpdateID', self.update_id)
            #if parent:
            #    value = (parent.get_id(),parent.get_update_id())
            #    if self.server:
            #        self.server.content_directory_server.set_variable(0, 'ContainerUpdateIDs', value)

        if mimetype == 'directory':
            return self.store[id]

        return None

    def len(self):
        return len(self.store)

    def get_by_id(self, id):
        if isinstance(id, basestring):
            id = id.split('@', 1)
            id = id[0]
        try:
            id = int(id)
        except ValueError:
            id = 1000

        if id == 0:
            id = 1000
        try:
            return self.store[id]
        except:
            return None

    def getnextID(self):
        self.next_id += 1
        return self.next_id

    def get_url_for_image(self, gallery2_id):
        url = self.gallery2_remote.get_URL_for_image(gallery2_id)
        return url

    def upnp_init(self):
        self.current_connection_id = None
        if self.server:
            self.server.connection_manager_server.set_variable(0, 'SourceProtocolInfo',
                                                                  'http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_TN;DLNA.ORG_OP=01;DLNA.ORG_FLAGS=00f00000000000000000000000000000,'
                                                                  'http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_SM;DLNA.ORG_OP=01;DLNA.ORG_FLAGS=00f00000000000000000000000000000,'
                                                                  'http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_MED;DLNA.ORG_OP=01;DLNA.ORG_FLAGS=00f00000000000000000000000000000,'
                                                                  'http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_LRG;DLNA.ORG_OP=01;DLNA.ORG_FLAGS=00f00000000000000000000000000000,'
                                                                  'http-get:*:image/jpeg:*,'
                                                                  'http-get:*:image/gif:*,'
                                                                  'http-get:*:image/png:*',
                                                                default=True)

    def retrieveAlbums(self, album_gallery2_id, parent):
        d = self.gallery2_remote.fetch_albums()

        def gotAlbums (albums):
            if albums:
                albums = [album for album in albums.values() if album.get('parent') == album_gallery2_id]
                if album_gallery2_id == '0' and len(albums) == 1:
                    album = albums[0]
                    self.store[1000].gallery2_id = album.get('name')
                    self.store[1000].name = album.get('title')
                    self.store[1000].description = album.get('summary')
                else:
                    for album in albums:
                        gallery2_id = album.get('name')
                        parent_gallery2_id = album.get('parent')
                        title = album.get('title')
                        description = album.get('summary')
                        store_item = {
                                  'name': id,
                                  'gallery2_id': gallery2_id,
                                  'parent_id': parent_gallery2_id,
                                  'title': title,
                                  'description': description,
                                  'mimetype': 'directory',
                                    }
                        self.append(store_item, parent)

        d.addCallback(gotAlbums)
        return d

    def retrieveItemsForAlbum (self, album_id, parent):
        # retrieve subalbums
        d1 = self.retrieveAlbums(album_id, parent)

        # retrieve images
        d2 = self.gallery2_remote.fetch_album_images(album_id)

        def gotImages(images):
            if images:
                for image in images:
                    image_gallery2_id = image.get('name')
                    parent_gallery2_id = image.get('parent')
                    thumbnail_gallery2_id = image.get('thumbName')
                    resized_gallery2_id = image.get('resizedName')
                    title = image.get('title')
                    description = image.get('description')

                    gallery2_id = resized_gallery2_id
                    if gallery2_id == '':
                        gallery2_id = image_gallery2_id

                    store_item = {
                                  'name': id,
                                  'gallery2_id': gallery2_id,
                                  'parent_id': parent_gallery2_id,
                                  'thumbnail_gallery2_id': thumbnail_gallery2_id,
                                  'title': title,
                                  'description': description,
                                  'mimetype': 'image/jpeg',
                                }
                    self.append(store_item, parent)

        d2.addCallback(gotImages)
        dl = defer.DeferredList([d1, d2])
        return dl


def main():

    f = Gallery2Store(None)

    def got_upnp_result(result):
        print "upnp", result

    f.upnp_init()


if __name__ == '__main__':

    from twisted.internet import reactor

    reactor.callWhenRunning(main)
    reactor.run()



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2010 Frank Scholz <dev@coherence-project.org>

from coherence.upnp.devices.wan_device_client import WANDeviceClient

from coherence import log

import coherence.extern.louie as louie


class InternetGatewayDeviceClient(log.Loggable):
    logCategory = 'igd_client'

    def __init__(self, device):
        log.Loggable.__init__(self)
        self.device = device
        self.device_type = self.device.get_friendly_device_type()
        self.version = int(self.device.get_device_type_version())
        self.icons = device.icons

        self.wan_device = None

        self.detection_completed = False

        louie.connect(self.embedded_device_notified, signal='Coherence.UPnP.EmbeddedDeviceClient.detection_completed', sender=self.device)

        try:
            wan_device = self.device.get_embedded_device_by_type('WANDevice')[0]
            self.wan_device = WANDeviceClient(wan_device)
        except:
            self.warning("Embedded WANDevice device not available, device not implemented properly according to the UPnP specification")
            raise

        self.info("InternetGatewayDevice %s", self.device.get_friendly_name())

    def remove(self):
        self.info("removal of InternetGatewayDeviceClient started")
        if self.wan_device != None:
            self.wan_device.remove()

    def embedded_device_notified(self, device):
        self.info("EmbeddedDevice %r sent notification", device)
        if self.detection_completed == True:
            return
        self.detection_completed = True
        louie.send('Coherence.UPnP.DeviceClient.detection_completed', None,
                               client=self, udn=self.device.udn)



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# <EXPERIMENTAL>

# a backend to expose to the local network the content of an audio CD
# inserted in a local drive
# the CD data is extracted from the CDDB/FreeDB database

# Warning: this backend does not detect insertion and ejection of CD drive
# the CD must be inserted before activating the backing.
# The backend must be disabled before ejecting the audio CD.

# Dependencies
# CDDB.py and DiscID.py (ex: debian package python-cddb)
# python-gst (with base and ugly plugins)

# TODO: switch from CDDB/FreeDB to musicBrainz
# TODO: find source for AlbumArt UI
# TODO: support other character encoding environment than ISO-8856-1

# Copyright 2010, Jean-Michel Sizun

import CDDB
import DiscID

from twisted.internet import reactor, threads

from coherence.upnp.core import DIDLLite
from coherence import log

from coherence.transcoder import GStreamerPipeline

from coherence.backend import AbstractBackendStore, Container, BackendItem

PLAY_TRACK_GST_PIPELINE = "cdiocddasrc device=%s track=%d ! wavenc name=enc"
TRACK_MIMETYPE = "audio/x-wav"
TRACK_FOURTH_FIELD = "*"


class TrackItem(BackendItem):
    logCategory = "audiocd"

    def __init__(self, device_name="/dev/cdrom", track_number=1, artist="Unknown", title="Unknown"):
        BackendItem.__init__(self)
        self.device_name = device_name
        self.track_number = track_number
        self.artist = artist
        self.title = title
        self.mimetype = TRACK_MIMETYPE
        self.fourth_field = TRACK_FOURTH_FIELD
        self.item = None
        self.pipeline = PLAY_TRACK_GST_PIPELINE % (self.device_name, self.track_number)
        self.location = GStreamerPipeline(self.pipeline, self.mimetype)

    def get_item(self):
        if self.item == None:
            upnp_id = self.storage_id
            upnp_parent_id = self.parent.get_id()
            url = self.store.urlbase + str(self.storage_id)
            self.item = DIDLLite.MusicTrack(upnp_id, upnp_parent_id, self.title)

            res = DIDLLite.Resource(url, 'http-get:*:%s:%s' % (self.mimetype, self.fourth_field))
            #res.duration = self.duration
            #res.size = self.get_size()
            self.item.res.append(res)
        return self.item

    def get_name(self):
        return self.title

    def get_path(self):
        return self.location

    def get_size(self):
        return self.size

    def get_id (self):
        return self.storage_id


class AudioCDStore(AbstractBackendStore):

    logCategory = 'audiocd'

    implements = ['MediaServer']

    description = ('audioCD', '', None)

    options = [{'option': 'version', 'text': 'UPnP Version:', 'type': 'int', 'default': 2, 'enum': (2, 1), 'help': 'the highest UPnP version this MediaServer shall support', 'level': 'advance'},
       {'option': 'uuid', 'text': 'UUID Identifier:', 'type': 'string', 'help': 'the unique (UPnP) identifier for this MediaServer, usually automatically set', 'level': 'advance'},
       {'option': 'device_name', 'text': 'device name for audio CD:', 'type': 'string', 'help': 'device name containing the audio cd.'}
    ]

    disc_title = None
    cdrom = None

    def __init__(self, server, **kwargs):
        AbstractBackendStore.__init__(self, server, **kwargs)

        self.name = 'audio CD'
        self.device_name = kwargs.get('device_name', "/dev/cdom")

        threads.deferToThread(self.extractAudioCdInfo)
        # self.init_completed() # will be fired when the audio CD info is extracted

    def upnp_init(self):
        self.current_connection_id = None
        if self.server:
            self.server.connection_manager_server.set_variable(0, 'SourceProtocolInfo',
                        ['http-get:*:%s:%s' % (TRACK_MIMETYPE, TRACK_FOURTH_FIELD)],
                        default=True)
            self.server.content_directory_server.set_variable(0, 'SystemUpdateID', self.update_id)
            #self.server.content_directory_server.set_variable(0, 'SortCapabilities', '*')

    def extractAudioCdInfo (self):
        """ extract the CD info (album art + artist + tracks), and construct the UPnP items"""
        self.cdrom = DiscID.open(self.device_name)
        disc_id = DiscID.disc_id(self.cdrom)

        (query_status, query_info) = CDDB.query(disc_id)
        if query_status in (210, 211):
            query_info = query_info[0]
        (read_status, read_info) = CDDB.read(query_info['category'], query_info['disc_id'])
#        print query_info['title']
#        print disc_id[1]
#        for i in range(disc_id[1]):
#            print "Track %.02d: %s" % (i, read_info['TTITLE' + `i`])

        track_count = disc_id[1]
        disc_id = query_info['disc_id']
        self.disc_title = query_info['title'].encode('utf-8')
        tracks = {}
        for i in range(track_count):
            tracks[i + 1] = read_info['TTITLE' + `i`].decode('ISO-8859-1').encode('utf-8')

        self.name = self.disc_title

        root_item = Container(None, self.disc_title)

        # we will sort the item by "track_number"
        def childs_sort(x, y):
            return cmp(x.track_number, y.track_number)
        root_item.sorting_method = childs_sort

        self.set_root_item(root_item)

        for number, title in tracks.items():
            item = TrackItem(self.device_name, number, "Unknown", title)
            external_id = "%s_%d" % (disc_id, number)
            root_item.add_child(item, external_id=external_id)

        self.info('Sharing audio CD %s', self.disc_title)

        reactor.callLater(2, self.checkIfAudioCdStillPresent)
        self.init_completed()

    def  checkIfAudioCdStillPresent(self):
        try:
            disc_id = DiscID.disc_id(self.cdrom)
            reactor.callLater(2, self.checkIfAudioCdStillPresent)
        except:
            self.warning('audio CD %s ejected: closing UPnP server!', self.disc_title)
            self.server.coherence.remove_plugin(self.server)

    def __repr__(self):
        return self.__class__.__name__



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2009, Frank Scholz <coherence@beebits.net>

# ScheduledRecording service

from twisted.web import resource

from coherence.upnp.core.soap_service import UPnPPublisher

from coherence.upnp.core import service


class ScheduledRecordingControl(service.ServiceControl, UPnPPublisher):

    def __init__(self, server):
        service.ServiceControl.__init__(self)
        UPnPPublisher.__init__(self)
        self.service = server
        self.variables = server.get_variables()
        self.actions = server.get_actions()


class ScheduledRecordingServer(service.ServiceServer, resource.Resource):

    implementation = 'optional'

    def __init__(self, device, backend=None):
        self.device = device
        if backend == None:
            backend = self.device.backend
        resource.Resource.__init__(self)
        self.version = 1
        service.ServiceServer.__init__(self, 'ScheduledRecording', self.version, backend)

        self.control = ScheduledRecordingControl(self)
        self.putChild(self.scpd_url, service.scpdXML(self, self.control))
        self.putChild(self.control_url, self.control)

    def listchilds(self, uri):
        cl = ''
        for c in self.children:
            cl += '<li><a href=%s/%s>%s</a></li>' % (uri, c, c)
        return cl

    def render(self, request):
        return '<html><p>root of the ScheduledRecording</p><p><ul>%s</ul></p></html>' % self.listchilds(request.uri)




# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2006-2008, Frank Scholz <coherence@beebits.net>

class ConnectionManagerClient:

    def __init__(self, service):
        self.service = service
        self.namespace = service.get_type()
        self.url = service.get_control_url()
        self.service.client = self
        self.service.subscribe()
    #def __del__(self):
    #    #print "ConnectionManagerClient deleted"
    #    pass

    def connection_manager_id(self):
        return "%s/%s" % (self.service.device.get_id(), self.service.get_id())

    def remove(self):
        self.service.remove()
        self.service = None
        self.namespace = None
        self.url = None
        del self

    def subscribe_for_variable(self, var_name, callback, signal=False):
        self.service.subscribe_for_variable(var_name, instance=0, callback=callback, signal=signal)

    def get_protocol_info(self):
        action = self.service.get_action('GetProtocolInfo')
        return action.call()

    def prepare_for_connection(self, remote_protocol_info, peer_connection_manager, peer_connection_id, direction):
        action = self.service.get_action('PrepareForConnection')
        if action:  # optional
            return action.call(RemoteProtocolInfo=remote_protocol_info,
                            PeerConnectionManager=peer_connection_manager,
                            PeerConnectionID=peer_connection_id,
                            Direction=direction)
        return None

    def connection_complete(self, connection_id):
        action = self.service.get_action('ConnectionComplete')
        if action:  # optional
            return action.call(ConnectionID=connection_id)
        return None

    def get_current_connection_ids(self):
        action = self.service.get_action('GetCurrentConnectionIDs')
        return action.call()

    def get_current_connection_info(self, connection_id):
        action = self.service.get_action('GetCurrentConnectionInfo')
        return action.call(ConnectionID=connection_id)



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2005, Tim Potter <tpot@samba.org>
# Copyright 2006 John-Mark Gurney <gurney_j@resnet.uroegon.edu>
# Copyright (C) 2006 Fluendo, S.A. (www.fluendo.com).
# Copyright 2006,2007,2008,2009 Frank Scholz <coherence@beebits.net>
#
# Implementation of a SSDP server under Twisted Python.
#

import random
import string
import time
import socket

from twisted.internet.protocol import DatagramProtocol
from twisted.internet import reactor, error
from twisted.internet import task
from twisted.web.http import datetimeToString
from coherence import log, SERVER_ID
import coherence.extern.louie as louie


SSDP_PORT = 1900
SSDP_ADDR = '239.255.255.250'


class SSDPServer(DatagramProtocol, log.Loggable):
    """A class implementing a SSDP server.  The notifyReceived and
    searchReceived methods are called when the appropriate type of
    datagram is received by the server."""
    logCategory = 'ssdp'

    def __init__(self, test=False, interface=''):
        # Create SSDP server
        log.Loggable.__init__(self)
        self.known = {}
        self._callbacks = {}
        self.test = test
        if not self.test:
          self.port = reactor.listenMulticast(SSDP_PORT, self, listenMultiple=True, interface=interface)

          self.port.joinGroup(SSDP_ADDR, interface=interface)

          self.resend_notify_loop = task.LoopingCall(self.resendNotify)
          self.resend_notify_loop.start(777.0, now=False)

          self.check_valid_loop = task.LoopingCall(self.check_valid)
          self.check_valid_loop.start(333.0, now=False)

        self.active_calls = []

    def shutdown(self):
        for call in reactor.getDelayedCalls():
            if call.func == self.send_it:
                call.cancel()
        if not self.test:
            if self.resend_notify_loop.running:
                self.resend_notify_loop.stop()
            if self.check_valid_loop.running:
                self.check_valid_loop.stop()
            '''Make sure we send out the byebye notifications.'''
            for st in self.known:
                if self.known[st]['MANIFESTATION'] == 'local':
                    self.doByebye(st)

    def datagramReceived(self, data, (host, port)):
        """Handle a received multicast datagram."""

        try:
            header, payload = data.split('\r\n\r\n')[:2]
        except ValueError, err:
            print err
            print 'Arggg,', data
            import pdb; pdb.set_trace()

        lines = header.split('\r\n')
        cmd = string.split(lines[0], ' ')
        lines = map(lambda x: x.replace(': ', ':', 1), lines[1:])
        lines = filter(lambda x: len(x) > 0, lines)

        headers = [string.split(x, ':', 1) for x in lines]
        headers = dict(map(lambda x: (x[0].lower(), x[1]), headers))

        self.msg('SSDP command %s %s - from %s:%d', cmd[0], cmd[1], host, port)
        self.debug('with headers: %s', headers)
        if cmd[0] == 'M-SEARCH' and cmd[1] == '*':
            # SSDP discovery
            self.discoveryRequest(headers, (host, port))
        elif cmd[0] == 'NOTIFY' and cmd[1] == '*':
            # SSDP presence
            self.notifyReceived(headers, (host, port))
        else:
            self.warning('Unknown SSDP command %s %s', cmd[0], cmd[1])

        # make raw data available
        # send out the signal after we had a chance to register the device
        louie.send('UPnP.SSDP.datagram_received', None, data, host, port)

    def register(self, manifestation, usn, st, location,
                        server=SERVER_ID,
                        cache_control='max-age=1800',
                        silent=False,
                        host=None):
        """Register a service or device that this SSDP server will
        respond to."""

        self.info('Registering %s (%s)', st, location)

        self.known[usn] = {}
        self.known[usn]['USN'] = usn
        self.known[usn]['LOCATION'] = location
        self.known[usn]['ST'] = st
        self.known[usn]['EXT'] = ''
        self.known[usn]['SERVER'] = server
        self.known[usn]['CACHE-CONTROL'] = cache_control

        self.known[usn]['MANIFESTATION'] = manifestation
        self.known[usn]['SILENT'] = silent
        self.known[usn]['HOST'] = host
        self.known[usn]['last-seen'] = time.time()

        self.msg(self.known[usn])

        if manifestation == 'local':
            self.doNotify(usn)

        if st == 'upnp:rootdevice':
            louie.send('Coherence.UPnP.SSDP.new_device', None, device_type=st, infos=self.known[usn])
            #self.callback("new_device", st, self.known[usn])

    def unRegister(self, usn):
        self.msg("Un-registering %s", usn)
        st = self.known[usn]['ST']
        if st == 'upnp:rootdevice':
            louie.send('Coherence.UPnP.SSDP.removed_device', None, device_type=st, infos=self.known[usn])
            #self.callback("removed_device", st, self.known[usn])

        del self.known[usn]

    def isKnown(self, usn):
        return self.known.has_key(usn)

    def notifyReceived(self, headers, (host, port)):
        """Process a presence announcement.  We just remember the
        details of the SSDP service announced."""

        self.info('Notification from (%s,%d) for %s', host, port, headers['nt'])
        self.debug('Notification headers: %s', headers)

        if headers['nts'] == 'ssdp:alive':
            try:
                self.known[headers['usn']]['last-seen'] = time.time()
                self.debug('updating last-seen for %r', headers['usn'])
            except KeyError:
                self.register('remote', headers['usn'], headers['nt'], headers['location'],
                              headers['server'], headers['cache-control'], host=host)
        elif headers['nts'] == 'ssdp:byebye':
            if self.isKnown(headers['usn']):
                self.unRegister(headers['usn'])
        else:
            self.warning('Unknown subtype %s for notification type %s',
                    headers['nts'], headers['nt'])
        louie.send('Coherence.UPnP.Log', None, 'SSDP', host, 'Notify %s for %s' % (headers['nts'], headers['usn']))

    def send_it(self, response, destination, delay, usn):
        self.info('send discovery response delayed by %ds for %s to %r', delay, usn, destination)
        try:
            self.transport.write(response, destination)
        except (AttributeError, socket.error), msg:
            self.info("failure sending out byebye notification: %r", msg)

    def discoveryRequest(self, headers, (host, port)):
        """Process a discovery request.  The response must be sent to
        the address specified by (host, port)."""

        self.info('Discovery request from (%s,%d) for %s', host, port, headers['st'])
        self.info('Discovery request for %s', headers['st'])

        louie.send('Coherence.UPnP.Log', None, 'SSDP', host, 'M-Search for %s' % headers['st'])

        # Do we know about this service?
        for i in self.known.values():
            if i['MANIFESTATION'] == 'remote':
                continue
            if(headers['st'] == 'ssdp:all' and
               i['SILENT'] == True):
                continue
            if(i['ST'] == headers['st'] or
                headers['st'] == 'ssdp:all'):
                response = []
                response.append('HTTP/1.1 200 OK')

                for k, v in i.items():
                    if k == 'USN':
                        usn = v
                    if k not in ('MANIFESTATION', 'SILENT', 'HOST'):
                        response.append('%s: %s' % (k, v))
                response.append('DATE: %s' % datetimeToString())

                response.extend(('', ''))
                delay = random.randint(0, int(headers['mx']))

                reactor.callLater(delay, self.send_it,
                                '\r\n'.join(response), (host, port), delay, usn)

    def doNotify(self, usn):
        """Do notification"""

        if self.known[usn]['SILENT'] == True:
            return
        self.info('Sending alive notification for %s', usn)

        resp = ['NOTIFY * HTTP/1.1',
            'HOST: %s:%d' % (SSDP_ADDR, SSDP_PORT),
            'NTS: ssdp:alive',
            ]
        stcpy = dict(self.known[usn].iteritems())
        stcpy['NT'] = stcpy['ST']
        del stcpy['ST']
        del stcpy['MANIFESTATION']
        del stcpy['SILENT']
        del stcpy['HOST']
        del stcpy['last-seen']

        resp.extend(map(lambda x: ': '.join(x), stcpy.iteritems()))
        resp.extend(('', ''))
        self.debug('doNotify content %s', resp)
        try:
            self.transport.write('\r\n'.join(resp), (SSDP_ADDR, SSDP_PORT))
            self.transport.write('\r\n'.join(resp), (SSDP_ADDR, SSDP_PORT))
        except (AttributeError, socket.error), msg:
            self.info("failure sending out alive notification: %r", msg)

    def doByebye(self, usn):
        """Do byebye"""

        self.info('Sending byebye notification for %s', usn)

        resp = ['NOTIFY * HTTP/1.1',
                'HOST: %s:%d' % (SSDP_ADDR, SSDP_PORT),
                'NTS: ssdp:byebye',
                ]
        try:
            stcpy = dict(self.known[usn].iteritems())
            stcpy['NT'] = stcpy['ST']
            del stcpy['ST']
            del stcpy['MANIFESTATION']
            del stcpy['SILENT']
            del stcpy['HOST']
            del stcpy['last-seen']
            resp.extend(map(lambda x: ': '.join(x), stcpy.iteritems()))
            resp.extend(('', ''))
            self.debug('doByebye content %s', resp)
            if self.transport:
                try:
                    self.transport.write('\r\n'.join(resp), (SSDP_ADDR, SSDP_PORT))
                except (AttributeError, socket.error), msg:
                    self.info("failure sending out byebye notification: %r", msg)
        except KeyError, msg:
            self.debug("error building byebye notification: %r", msg)

    def resendNotify(self):
        for usn in self.known:
            if self.known[usn]['MANIFESTATION'] == 'local':
                self.doNotify(usn)

    def check_valid(self):
        """ check if the discovered devices are still ok, or
            if we haven't received a new discovery response
        """
        self.debug("Checking devices/services are still valid")
        removable = []
        for usn in self.known:
            if self.known[usn]['MANIFESTATION'] != 'local':
                _, expiry = self.known[usn]['CACHE-CONTROL'].split('=')
                expiry = int(expiry)
                now = time.time()
                last_seen = self.known[usn]['last-seen']
                self.debug("Checking if %r is still valid - last seen %d (+%d), now %d", self.known[usn]['USN'], last_seen, expiry, now)
                if last_seen + expiry + 30 < now:
                    self.debug("Expiring: %r", self.known[usn])
                    if self.known[usn]['ST'] == 'upnp:rootdevice':
                        louie.send('Coherence.UPnP.SSDP.removed_device', None, device_type=self.known[usn]['ST'], infos=self.known[usn])
                    removable.append(usn)
        while len(removable) > 0:
            usn = removable.pop(0)
            del self.known[usn]

    def subscribe(self, name, callback):
        self._callbacks.setdefault(name, []).append(callback)

    def unsubscribe(self, name, callback):
        callbacks = self._callbacks.get(name, [])
        if callback in callbacks:
            callbacks.remove(callback)
        self._callbacks[name] = callbacks

    def callback(self, name, *args):
        for callback in self._callbacks.get(name, []):
            callback(*args)



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2006-2008, Frank Scholz <coherence@beebits.net>

from coherence import log


class AVTransportClient(log.Loggable):
    logCategory = 'avtransportclient'

    def __init__(self, service):
        log.Loggable.__init__(self)
        self.service = service
        self.namespace = service.get_type()
        self.url = service.get_control_url()
        self.service.subscribe()
        self.service.client = self
    #def __del__(self):
    #    #print "AVTransportClient deleted"
    #    pass

    def remove(self):
        self.service.remove()
        self.service = None
        self.namespace = None
        self.url = None
        del self

    def subscribe_for_variable(self, var_name, callback, signal=False):
        self.service.subscribe_for_variable(var_name, instance=0, callback=callback, signal=signal)

    def set_av_transport_uri(self, instance_id=0, current_uri='', current_uri_metadata=''):
        action = self.service.get_action('SetAVTransportURI')
        return action.call(InstanceID=instance_id,
                            CurrentURI=current_uri,
                            CurrentURIMetaData=current_uri_metadata)

    def set_next_av_transport_uri(self, instance_id=0, next_uri='', next_uri_metadata=''):
        action = self.service.get_action('SetNextAVTransportURI')
        if action:  # optional
            return action.call(InstanceID=instance_id,
                            NextURI=next_uri,
                            NextURIMetaData=next_uri_metadata)
        return None

    def get_media_info(self, instance_id=0):
        action = self.service.get_action('GetMediaInfo')
        return action.call(InstanceID=instance_id)

    def get_media_info_ext(self, instance_id=0):
        action = self.service.get_action('GetMediaInfo_Ext')
        return action.call(InstanceID=instance_id)

    def get_transport_info(self, instance_id=0):
        action = self.service.get_action('GetTransportInfo')
        return action.call(InstanceID=instance_id)

    def get_position_info(self, instance_id=0):
        action = self.service.get_action('GetPositionInfo')
        return action.call(InstanceID=instance_id)

    def get_device_capabilities(self, instance_id=0):
        action = self.service.get_action('GetDeviceCapabilities')
        return action.call(InstanceID=instance_id)

    def get_transport_settings(self, instance_id=0):
        action = self.service.get_action('GetTransportSettings')
        return action.call(InstanceID=instance_id)

    def pause(self, instance_id=0):
        action = self.service.get_action('Pause')
        if action:  # optional
            return action.call(InstanceID=instance_id)
        return None

    def play(self, instance_id=0, speed=1):
        action = self.service.get_action('Play')
        return action.call(InstanceID=instance_id, Speed=speed)

    def stop(self, instance_id=0):
        action = self.service.get_action('Stop')
        return action.call(InstanceID=instance_id)

    def record(self, instance_id=0):
        action = self.service.get_action('Record')
        if action:  # optional
            return action.call(InstanceID=instance_id)
        return None

    def seek(self, instance_id=0, unit='', target=0):
        action = self.service.get_action('Seek')
        return action.call(InstanceID=instance_id,
                            Unit=unit,
                            Target=target)

    def next(self, instance_id=0):
        action = self.service.get_action('Next')
        return action.call(InstanceID=instance_id)

    def previous(self, instance_id=0):
        action = self.service.get_action('Previous')
        return action.call(InstanceID=instance_id)

    def get_current_transport_actions(self, instance_id=0):
        action = self.service.get_action('GetCurrentTransportActions')
        return action.call(InstanceID=instance_id)



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2015, https://github.com/unintended

"""
A backend to access twitch.tv streams.

To enable personalized features (e.g. 'Following' streams), add 'access_token' key into your config file.

1. To get access token copy and paste the following into your browser:
  https://api.twitch.tv/kraken/oauth2/authorize?response_type=token&client_id=37684tuwyxmogmtduz6lz0jdtf0acob&redirect_uri=http://localhost&scope=user_read
2. After authorization you will be redirected to http://localhost with access token in fragment part, e.g.
  http://localhost/#access_token=<YOUR_ACCESS_TOKEN_IS_HERE>&scope=user_read
3. Copy the token and paste in TwitchStore section of your config file:
  access_token = <YOUR_ACCESS_TOKEN (step 2)>
"""

import json
import urllib

from coherence.log import Loggable
from dateutil import parser as dateutil_parser
import livestreamer
from coherence.backend import AbstractBackendStore, Container, BackendItem, LazyContainer
from coherence.upnp.core import utils, DIDLLite
from twisted.internet import threads
from twisted.python.failure import Failure
from twisted.web import server, http
from twisted.web.resource import Resource
from twisted.web.static import NoRangeStaticProducer


MPEG_MIME = 'video/mpeg'

TWITCH_API_URL = 'https://api.twitch.tv/kraken'


class LiveStreamerProxyResource(Resource, Loggable):
  logCategory = 'twitch_store'

  def __init__(self, url, stream_id, content_type=MPEG_MIME):
    Resource.__init__(self)
    Loggable.__init__(self)
    self.url = url
    self.stream_id = stream_id
    self.content_type = content_type

  def render_GET(self, request):
    self.debug("serving %s request from %s for %s", request.method, request.getClientIP(), request.uri)

    def stream_opened(fd):
      producer = NoRangeStaticProducer(request, fd)
      producer.start()

    def got_streams(streams):
      if self.stream_id not in streams:
        self.warning("stream not found for %s@%s", self.url, self.stream_id)
        request.setResponseCode(http.NOT_FOUND)
        request.write('')
        return

      request.setHeader('Content-Type', self.content_type)
      request.setResponseCode(http.OK)

      if request.method == 'HEAD':
        request.write('')
        return

      d_open_stream = threads.deferToThread(streams[self.stream_id].open)
      d_open_stream.addCallback(stream_opened)

    d_get_streams = threads.deferToThread(livestreamer.streams, self.url)
    d_get_streams.addCallback(got_streams)

    return server.NOT_DONE_YET


class TwitchLazyContainer(LazyContainer):
  logCategory = 'twitch_store'

  def __init__(self, parent, title, limit=None, **kwargs):
    super(TwitchLazyContainer, self).__init__(parent, title, **kwargs)

    self.childrenRetriever = self._retrieve_children
    self.refresh = 60
    self.children_url = None
    self.limit = limit

  def result_handler(self, result, **kwargs):
    return True

  def _retrieve_children(self, parent=None, **kwargs):
    if self.children_url is None:
      return

    kwargs.update({'limit': self.limit})
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    url = "%s?%s" % (self.children_url, urllib.urlencode(kwargs)) if kwargs else self.children_url

    d = utils.getPage(url)
    d.addCallbacks(self._got_page, self._got_error)
    return d

  def _got_page(self, result):
    self.info('connection to twitch service successful for game list')
    result = json_loads(result)
    return self.result_handler(result)

  def _got_error(self, error):
    self.warning("connection to twitch.tv service failed: %s", self.children_url)
    self.debug("%r", error.getTraceback())
    self.childrenRetrievingNeeded = True  # we retry
    return Failure("Unable to retrieve game list")


class GamesContainer(TwitchLazyContainer):
  def __init__(self, parent, title='Games', description=None, limit=None, children_limit=None, **kwargs):
    super(GamesContainer, self).__init__(parent, title, limit=limit, **kwargs)
    self.description = description

    self.children_url = '%s/games/top' % TWITCH_API_URL
    self.sorting_method = sort_by_viewers
    self.children_limit = children_limit

  def result_handler(self, result, **kwargs):
    for game_info in result['top']:
      game_name = game_info['game']['name']
      item = StreamsContainer(self, game_name,
                              viewers=game_info['viewers'],
                              channels=game_info['channels'],
                              cover_url=game_info['game']['box']['large'],
                              game=game_name,
                              limit=self.children_limit)
      # item.description = "%d viewers" % game_info['viewers']
      self.add_child(item, external_id=game_info['game']['_id'])
    return True


class StreamsContainer(TwitchLazyContainer):
  URL = '%s/streams/'

  def __init__(self, parent, title, viewers=0, channels=0, streams_url=URL, cover_url=None, **kwargs):
    super(StreamsContainer, self).__init__(parent, title, **kwargs)
    self.viewers = viewers
    self.channels = channels

    self.children_url = streams_url % TWITCH_API_URL
    self.cover_url = cover_url
    self.sorting_method = sort_by_viewers

  def result_handler(self, result, **kwargs):
    for stream in result['streams']:
      created_at = dateutil_parser.parse(stream['created_at'])
      item = TwitchStreamItem(stream['channel']['display_name'], stream['channel']['url'],
                              status=stream['channel']['status'],
                              viewers=stream['viewers'],
                              preview_url=stream['preview']['medium'],
                              created_at=created_at)
      self.add_child(item, external_id='stream%d' % stream['_id'])
    return True


class TwitchStreamItem(BackendItem):
  logCategory = 'twitch_store'

  def __init__(self, title, url, status=None, viewers=0, created_at=None, preview_url=None):
    BackendItem.__init__(self)
    self.name = title
    self.status = status
    self.mimetype = MPEG_MIME
    self.created_at = created_at
    self.preview_url = preview_url
    self.viewers = viewers
    self.location = LiveStreamerProxyResource(url, 'best')

  def get_item(self):
    if self.item is None:
      upnp_id = self.get_id()
      upnp_parent_id = self.parent.get_id()

      self.item = DIDLLite.VideoItem(upnp_id, upnp_parent_id, self.name)
      self.item.description = self.status
      self.item.longDescription = self.status
      self.item.date = self.created_at
      self.item.albumArtURI = self.preview_url

      res = DIDLLite.Resource(self.url, 'http-get:*:%s:#' % MPEG_MIME)
      self.item.res.append(res)
    return self.item

  def get_id(self):
    return self.storage_id

  def get_url(self):
    return self.url

  def replace_by(self, item):
    # TODO update fields
    return True


class TwitchStore(AbstractBackendStore):
  logCategory = 'twitch_store'

  implements = ['MediaServer']

  wmc_mapping = {'16': 1000}

  description = ('twitch.tv', 'twitch.tv', None)

  options = [
    {
      'option': 'name',
      'text': 'Server Name:',
      'type': 'string',
      'default': 'twitch.tv',
      'help': 'the name under this MediaServer shall show up with on other UPnP clients'
    },
    {
      'option': 'access_token',
      'text': 'OAuth Access Token:',
      'type': 'string',
      'default': '',
      'help': 'access token to show personalized list of followed streams'
    },
    {
      'option': 'version',
      'text': 'UPnP Version:',
      'type': 'int',
      'default': 2,
      'enum': (2, 1),
      'help': 'the highest UPnP version this MediaServer shall support',
      'level': 'advance'
    },
    {
      'option': 'uuid',
      'text': 'UUID Identifier:',
      'type': 'string',
      'default': 'twitch_tv',
      'help': 'the unique (UPnP) identifier for this MediaServer',
      'level': 'advance'
    }
  ]

  def __init__(self, server, **kwargs):
    AbstractBackendStore.__init__(self, server, **kwargs)

    self.name = self.config.get('name', 'twitch.tv')
    self.uuid = self.config.get('uuid', 'twitch_tv')
    self.access_token = self.config.get('access_token')

    self.init_completed()

  def __repr__(self):
    return self.__class__.__name__

  def upnp_init(self):
    if self.server:
      self.server.connection_manager_server.set_variable(0, 'SourceProtocolInfo',
                                                         ['http-get:*:%s:*' % MPEG_MIME],
                                                         default=True)
    # root item
    root_item = Container(None, self.name)
    self.set_root_item(root_item)

    # 'Following' directory
    settings = self.config.get('Following', {})
    if self.access_token and settings.get('active') != 'no':
      games_dir = StreamsContainer(root_item,
                                   title=settings.get('name') or 'Following',
                                   streams_url='%s/streams/followed',
                                   limit=settings.get('limit', 25),
                                   oauth_token=self.access_token)
      root_item.add_child(games_dir)

    # 'Games' directory
    settings = self.config.get('TopGames', {})
    if settings.get('active') != 'no':
      games_dir = GamesContainer(root_item,
                                 title=settings.get('name', 'Top Games'),
                                 limit=settings.get('limit', 10),
                                 children_limit=settings.get('children_limit', 25))
      root_item.add_child(games_dir)

    # 'Top Streams' directory
    settings = self.config.get('TopStreams', {})
    if settings.get('active') != 'no':
      games_dir = StreamsContainer(root_item,
                                   title=settings.get('name', 'Top Streams'),
                                   limit=settings.get('limit', 25))
      root_item.add_child(games_dir)


def sort_by_viewers(x, y):
  return cmp(y.viewers, x.viewers)


def json_loads(data):
  if isinstance(data, (list, tuple)):
    data = data[0]
  return json.loads(data)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php
#
# Copyright 2009, Benjamin Kampmann <ben.kampmann@gmail.com>
# Copyright 2014, Hartmut Goebel <h.goebel@crazy-compilers.com>

from twisted.internet import reactor

from coherence.base import Coherence
from coherence.upnp.devices.control_point import ControlPoint
from coherence.upnp.core import DIDLLite


# browse callback
def process_media_server_browse(result, client):
  print "browsing root of", client.device.get_friendly_name()
  print "result contains", result['NumberReturned'],
  print "out of", result['TotalMatches'], "total matches."

  elt = DIDLLite.DIDLElement.fromString(result['Result'])
  for item in elt.getItems():

    if item.upnp_class.startswith("object.container"):
      print "  container", item.title, "(%s)" % item.id,
      print "with", item.childCount, "items."

    if item.upnp_class.startswith("object.item"):
      print "  item", item.title, "(%s)." % item.id


# called for each media server found
def media_server_found(client, udn):
  print "Media Server found:", client.device.get_friendly_name()

  d = client.content_directory.browse(0,
                                      browse_flag='BrowseDirectChildren',
                                      process_result=False,
                                      backward_compatibility=False)
  d.addCallback(process_media_server_browse, client)


# sadly they sometimes get removed as well :(
def media_server_removed(udn):
  print "Media Server gone:", udn


def start():
  control_point = ControlPoint(Coherence({'logmode': 'warning'}),
                               auto_client=['MediaServer'])
  control_point.connect(media_server_found, 'Coherence.UPnP.ControlPoint.MediaServer.detected')
  control_point.connect(media_server_removed, 'Coherence.UPnP.ControlPoint.MediaServer.removed')

  # now we should also try to discover the ones that are already there:
  for device in control_point.coherence.devices:
    print device

if __name__ == "__main__":
  reactor.callWhenRunning(start)
  reactor.run()



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2006, Frank Scholz <coherence@beebits.net>

import os

from twisted.internet import reactor
from twisted.python import filepath, util
from nevow import athena, inevow, loaders, tags, static
from twisted.web import server, resource

from zope.interface import implements, Interface

import coherence.extern.louie as louie

from coherence import log


class IWeb(Interface):

    def goingLive(self):
        pass


class Web(object):

    def __init__(self, coherence):
        super(Web, self).__init__()
        self.coherence = coherence


class MenuFragment(athena.LiveElement, log.Loggable):
    logCategory = 'webui_menu_fragment'
    jsClass = u'Coherence.Base'
    fragmentName = 'coherence-menu'

    docFactory = loaders.stan(
        tags.div(render=tags.directive('liveElement'))[
            tags.div(id="coherence_menu_box", class_="coherence_menu_box")[""],
        ]
        )

    def __init__(self, page):
        super(MenuFragment, self).__init__()
        self.setFragmentParent(page)
        self.page = page
        self.coherence = page.coherence
        self.tabs = []

    def going_live(self):
        self.info("add a view to the MenuFragment")
        d = self.page.notifyOnDisconnect()
        d.addCallback(self.remove_me)
        d.addErrback(self.remove_me)
        if len(self.tabs):
            return self.tabs
        else:
            return {}
    athena.expose(going_live)

    def add_tab(self, title, active, id):
        self.info("add tab %s to the MenuFragment", title)
        new_tab = {u'title': unicode(title),
                   u'active': unicode(active),
                   u'athenaid': u'athenaid:%d' % id}
        for t in self.tabs:
            if t[u'title'] == new_tab[u'title']:
                return
        self.tabs.append(new_tab)
        self.callRemote('addTab', new_tab)

    def remove_me(self, result):
        self.info("remove view from MenuFragment")


class DevicesFragment(athena.LiveElement, log.Loggable):
    logCategory = 'webui_device_fragment'
    jsClass = u'Coherence.Devices'
    fragmentName = 'coherence-devices'

    docFactory = loaders.stan(
        tags.div(render=tags.directive('liveElement'))[
            tags.div(id="Devices-container", class_="coherence_container")[""],
        ]
        )

    def __init__(self, page, active):
        super(DevicesFragment, self).__init__()
        self.setFragmentParent(page)
        self.page = page
        self.coherence = page.coherence
        self.active = active

    def going_live(self):
        self.info("add a view to the DevicesFragment %s", self._athenaID)
        self.page.menu.add_tab('Devices', self.active, self._athenaID)
        d = self.page.notifyOnDisconnect()
        d.addCallback(self.remove_me)
        d.addErrback(self.remove_me)
        devices = []
        for device in self.coherence.get_devices():
            if device is not None:
                devices.append({u'name': device.get_markup_name(),
                        u'usn': unicode(device.get_usn())})

        louie.connect(self.add_device,
                'Coherence.UPnP.Device.detection_completed', louie.Any)
        louie.connect(self.remove_device,
                'Coherence.UPnP.Device.removed', louie.Any)

        return devices

    athena.expose(going_live)

    def remove_me(self, result):
        self.info("remove view from the DevicesFragment")

    def add_device(self, device):
        self.info("DevicesFragment found device %s %s of type %s",
                                                device.get_usn(),
                                                device.get_friendly_name(),
                                                device.get_device_type())
        self.callRemote('addDevice',
                {u'name': device.get_markup_name(),
                u'usn': unicode(device.get_usn())})

    def remove_device(self, usn):
        self.info("DevicesFragment remove device %s", usn)
        self.callRemote('removeDevice', unicode(usn))

    def render_devices(self, ctx, data):
        cl = []
        self.info('children: %s', self.coherence.children)
        for child in self.coherence.children:
            device = self.coherence.get_device_with_id(child)
            if device is not None:
                cl.append(tags.li[tags.a(href='/' + child)[
                    device.get_friendly_device_type, ':',
                    device.get_device_type_version, ' ',
                    device.get_friendly_name()]])
            else:
                cl.append(tags.li[child])
        return ctx.tag[tags.ul[cl]]


class LoggingFragment(athena.LiveElement, log.Loggable):
    logCategory = 'webui_logging_fragment'
    jsClass = u'Coherence.Logging'
    fragmentName = 'coherence-logging'

    docFactory = loaders.stan(
        tags.div(render=tags.directive('liveElement'))[
            tags.div(id="Logging-container", class_="coherence_container")[""],
        ]
        )

    def __init__(self, page, active):
        super(LoggingFragment, self).__init__()
        self.setFragmentParent(page)
        self.page = page
        self.coherence = page.coherence
        self.active = active

    def going_live(self):
        self.info("add a view to the LoggingFragment %s", self._athenaID)
        self.page.menu.add_tab('Logging', self.active, self._athenaID)
        d = self.page.notifyOnDisconnect()
        d.addCallback(self.remove_me)
        d.addErrback(self.remove_me)
        return {}
    athena.expose(going_live)

    def remove_me(self, result):
        self.info("remove view from the LoggingFragment")


class WebUI(athena.LivePage, log.Loggable):
    """
    """
    logCategory = 'webui'
    jsClass = u'Coherence'

    addSlash = True
    docFactory = loaders.xmlstr("""\
<html xmlns:nevow="http://nevow.com/ns/nevow/0.1">
<head>
<nevow:invisible nevow:render="liveglue" />
<link rel="stylesheet" type="text/css" href="static/main.css" />
</head>
<body>
<div id="coherence_header"><div class="coherence_title">Coherence</div><div nevow:render="menu"></div></div>
<div id="coherence_body">
<div nevow:render="devices" />
<div nevow:render="logging" />
</div>
</body>
</html>
""")

    def __init__(self, *a, **kw):
        super(WebUI, self).__init__(*a, **kw)
        self.coherence = self.rootObject.coherence

        self.jsModules.mapping.update({
            'MochiKit': filepath.FilePath(__file__).parent().child('static').child('MochiKit.js').path})

        self.jsModules.mapping.update({
            'Coherence': filepath.FilePath(__file__).parent().child('static').child('Coherence.js').path})
        self.jsModules.mapping.update({
            'Coherence.Base': filepath.FilePath(__file__).parent().child('static').child('Coherence.Base.js').path})
        self.jsModules.mapping.update({
            'Coherence.Devices': filepath.FilePath(__file__).parent().child('static').child('Coherence.Devices.js').path})
        self.jsModules.mapping.update({
            'Coherence.Logging': filepath.FilePath(__file__).parent().child('static').child('Coherence.Logging.js').path})
        self.menu = MenuFragment(self)

    def childFactory(self, ctx, name):
        self.info('WebUI childFactory: %s', name)
        try:
            return self.rootObject.coherence.children[name]
        except:
            ch = super(WebUI, self).childFactory(ctx, name)
            if ch is None:
                p = util.sibpath(__file__, name)
                self.info('looking for file %s', p)
                if os.path.exists(p):
                    ch = static.File(p)
            return ch

    def render_listmenu(self, ctx, data):
        l = []
        l.append(tags.div(id="t", class_="coherence_menu_item")[tags.a(href='/' + 'devices', class_="coherence_menu_link")['Devices']])
        l.append(tags.div(id="t", class_="coherence_menu_item")[tags.a(href='/' + 'logging', class_="coherence_menu_link")['Logging']])
        return ctx.tag[l]

    def render_menu(self, ctx, data):
        self.info('render_menu')
        return self.menu

    def render_devices(self, ctx, data):
        self.info('render_devices')
        f = DevicesFragment(self, 'yes')
        return f

    def render_logging(self, ctx, data):
        self.info('render_logging')
        f = LoggingFragment(self, 'no')
        return f



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2010 Frank Scholz <dev@coherence-project.org>

from coherence.upnp.devices.wan_connection_device_client import WANConnectionDeviceClient

from coherence.upnp.services.clients.wan_common_interface_config_client import WANCommonInterfaceConfigClient

from coherence import log

import coherence.extern.louie as louie


class WANDeviceClient(log.Loggable):
    logCategory = 'wan_device_client'

    def __init__(self, device):
        log.Loggable.__init__(self)
        self.device = device
        self.device_type = self.device.get_friendly_device_type()
        self.version = int(self.device.get_device_type_version())
        self.icons = device.icons

        self.wan_connection_device = None
        self.wan_common_interface_connection = None

        self.embedded_device_detection_completed = False
        self.service_detection_completed = False

        louie.connect(self.embedded_device_notified, signal='Coherence.UPnP.EmbeddedDeviceClient.detection_completed', sender=self.device)

        try:
            wan_connection_device = self.device.get_embedded_device_by_type('WANConnectionDevice')[0]
            self.wan_connection_device = WANConnectionDeviceClient(wan_connection_device)
        except:
            self.warning("Embedded WANConnectionDevice device not available, device not implemented properly according to the UPnP specification")
            raise

        louie.connect(self.service_notified, signal='Coherence.UPnP.DeviceClient.Service.notified', sender=self.device)
        for service in self.device.get_services():
            if service.get_type() in ["urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1"]:
                self.wan_common_interface_connection = WANCommonInterfaceConfigClient(service)

        self.info("WANDevice %s", self.device.get_friendly_name())

    def remove(self):
        self.info("removal of WANDeviceClient started")
        if self.wan_common_interface_connection != None:
            self.wan_common_interface_connection.remove()
        if self.wan_connection_device != None:
            self.wan_connection_device.remove()

    def embedded_device_notified(self, device):
        self.info("EmbeddedDevice %r sent notification", device)
        if self.embedded_device_detection_completed == True:
            return

        self.embedded_device_detection_completed = True
        if self.embedded_device_detection_completed == True and self.service_detection_completed == True:
            louie.send('Coherence.UPnP.EmbeddedDeviceClient.detection_completed', None,
                               self)

    def service_notified(self, service):
        self.info("Service %r sent notification", service)
        if self.service_detection_completed == True:
            return
        if self.wan_common_interface_connection != None:
            if not hasattr(self.wan_common_interface_connection.service, 'last_time_updated'):
                return
            if self.wan_common_interface_connection.service.last_time_updated == None:
                return
        self.service_detection_completed = True
        if self.embedded_device_detection_completed == True and self.service_detection_completed == True:
            louie.send('Coherence.UPnP.EmbeddedDeviceClient.detection_completed', None,
                               self)



# -*- coding: utf-8 -*-

# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2008 Frank Scholz <coherence@beebits.net>

import os.path

from twisted.python import util
from twisted.web import resource, static
from twisted.internet import reactor

import coherence.extern.louie as louie

from coherence import log


class DeviceHttpRoot(resource.Resource, log.Loggable):
    logCategory = 'basicdevice'

    def __init__(self, server):
        resource.Resource.__init__(self)
        log.Loggable.__init__(self)
        self.server = server

    def getChildWithDefault(self, path, request):
        self.info('DeviceHttpRoot %s getChildWithDefault %s %s %s',
                  self.server.device_type, path, request.uri, request.client)
        self.info(request.getAllHeaders())
        if self.children.has_key(path):
            return self.children[path]
        if request.uri == '/':
            return self
        return self.getChild(path, request)

    def getChild(self, name, request):
        self.info('DeviceHttpRoot %s getChild %s', name, request)
        ch = None
        if ch is None:
            p = util.sibpath(__file__, name)
            if os.path.exists(p):
                ch = static.File(p)
        self.info('DeviceHttpRoot ch  %s', ch)
        return ch

    def listchilds(self, uri):
        cl = ''
        for c in self.children:
            cl += '<li><a href=%s/%s>%s</a></li>' % (uri, c, c)
        return cl

    def render(self, request):
        return '<html><p>root of the %s %s</p><p><ul>%s</ul></p></html>' % (self.server.backend.name,
                                                                           self.server.device_type,
                                                                           self.listchilds(request.uri))


# class RootDeviceXML(static.Data):
#   def __init__(self, hostname, uuid, urlbase,
#                xmlns='urn:schemas-upnp-org:device-1-0',
#                device_uri_base='urn:schemas-upnp-org:device',
#                device_type='BasicDevice',
#                version=2,
#                friendly_name='Coherence UPnP BasicDevice',
#                manufacturer='beebits.net',
#                manufacturer_url='http://coherence.beebits.net',
#                model_description='Coherence UPnP BasicDevice',
#                model_name='Coherence UPnP BasicDevice',
#                model_number=__version__,
#                model_url='http://coherence.beebits.net',
#                serial_number='0000001',
#                presentation_url='',
#                services=None,
#                devices=None,
#                icons=None,
#                dlna_caps=None):
#         uuid = str(uuid)
#         root = ET.Element('root')
#         root.attrib['xmlns'] = xmlns
#         device_type_uri = ':'.join((device_uri_base, device_type, str(version)))
#         e = ET.SubElement(root, 'specVersion')
#         ET.SubElement(e, 'major').text = '1'
#         ET.SubElement(e, 'minor').text = '0'
#         #ET.SubElement(root, 'URLBase').text = urlbase + uuid[5:] + '/'
#
#         d = ET.SubElement(root, 'device')
#
#         if device_type == 'MediaServer':
#             x = ET.SubElement(d, 'dev:X_DLNADOC')
#             x.text = 'DMS-1.50'
#             x = ET.SubElement(d, 'dev:X_DLNADOC')
#             x.text = 'M-DMS-1.50'
#         elif device_type == 'MediaRenderer':
#             x = ET.SubElement(d, 'dev:X_DLNADOC')
#             x.text = 'DMR-1.50'
#             x = ET.SubElement(d, 'dev:X_DLNADOC')
#             x.text = 'M-DMR-1.50'
#
#         if len(dlna_caps) > 0:
#             if isinstance(dlna_caps, basestring):
#                 dlna_caps = [dlna_caps]
#             for cap in dlna_caps:
#                 x = ET.SubElement(d, 'dev:X_DLNACAP')
#                 x.text = cap
#
#         ET.SubElement(d, 'deviceType').text = device_type_uri
#         ET.SubElement(d, 'friendlyName').text = friendly_name
#         ET.SubElement(d, 'manufacturer').text = manufacturer
#         ET.SubElement(d, 'manufacturerURL').text = manufacturer_url
#         ET.SubElement(d, 'modelDescription').text = model_description
#         ET.SubElement(d, 'modelName').text = model_name
#         ET.SubElement(d, 'modelNumber').text = model_number
#         ET.SubElement(d, 'modelURL').text = model_url
#         ET.SubElement(d, 'serialNumber').text = serial_number
#         ET.SubElement(d, 'UDN').text = uuid
#         ET.SubElement(d, 'UPC').text = ''
#         ET.SubElement(d, 'presentationURL').text = presentation_url
#
#         if len(services):
#             e = ET.SubElement(d, 'serviceList')
#             for service in services:
#                 id = service.get_id()
#                 s = ET.SubElement(e, 'service')
#                 try:
#                     namespace = service.namespace
#                 except:
#                     namespace = 'schemas-upnp-org'
#                 if(hasattr(service, 'version') and
#                     service.version < version):
#                     v = service.version
#                 else:
#                     v = version
#                 ET.SubElement(s, 'serviceType').text = 'urn:%s:service:%s:%d' % (namespace, id, int(v))
#                 try:
#                     namespace = service.id_namespace
#                 except:
#                     namespace = 'upnp-org'
#                 ET.SubElement(s, 'serviceId').text = 'urn:%s:serviceId:%s' % (namespace, id)
#                 ET.SubElement(s, 'SCPDURL').text = '/' + uuid[5:] + '/' + id + '/' + service.scpd_url
#                 ET.SubElement(s, 'controlURL').text = '/' + uuid[5:] + '/' + id + '/' + service.control_url
#                 ET.SubElement(s, 'eventSubURL').text = '/' + uuid[5:] + '/' + id + '/' + service.subscription_url
#
#         if len(devices):
#             e = ET.SubElement(d, 'deviceList')
#
#         if len(icons):
#             e = ET.SubElement(d, 'iconList')
#             for icon in icons:
#
#                 icon_path = ''
#                 if icon.has_key('url'):
#                     if icon['url'].startswith('file://'):
#                         icon_path = icon['url'][7:]
#                     elif icon['url'] == '.face':
#                         icon_path = os.path.join(os.path.expanduser('~'), ".face")
#                     else:
#                         from pkg_resources import resource_filename
#                         icon_path = os.path.abspath(resource_filename(__name__, os.path.join('..', '..', '..', 'misc', 'device-icons', icon['url'])))
#
#                 if os.path.exists(icon_path) == True:
#                     i = ET.SubElement(e, 'icon')
#                     for k, v in icon.items():
#                         if k == 'url':
#                             if v.startswith('file://'):
#                                 ET.SubElement(i, k).text = '/' + uuid[5:] + '/' + os.path.basename(v)
#                                 continue
#                             elif v == '.face':
#                                 ET.SubElement(i, k).text = '/' + uuid[5:] + '/' + 'face-icon.png'
#                                 continue
#                             else:
#                                 ET.SubElement(i, k).text = '/' + uuid[5:] + '/' + os.path.basename(v)
#                                 continue
#                         ET.SubElement(i, k).text = str(v)
#         #if self.has_level(LOG_DEBUG):
#         #    indent( root)
#
#         self.xml = """<?xml version="1.0" encoding="utf-8"?>""" + ET.tostring(root, encoding='utf-8')
#         static.Data.__init__(self, self.xml, 'text/xml')


class BasicDeviceMixin(object):

    def __init__(self, coherence, backend, **kwargs):
        self.coherence = coherence
        if not hasattr(self, 'version'):
            self.version = int(kwargs.get('version', self.coherence.config.get('version', 2)))

        try:
            self.uuid = kwargs['uuid']
            if not self.uuid.startswith('uuid:'):
                self.uuid = 'uuid:' + self.uuid
        except KeyError:
            from coherence.upnp.core.uuid import UUID
            self.uuid = UUID()

        self.backend = None
        urlbase = self.coherence.urlbase
        if urlbase[-1] != '/':
            urlbase += '/'
        self.urlbase = urlbase + str(self.uuid)[5:]

        kwargs['urlbase'] = self.urlbase
        self.icons = kwargs.get('iconlist', kwargs.get('icons', []))
        if len(self.icons) == 0:
            if kwargs.has_key('icon'):
                if isinstance(kwargs['icon'], dict):
                    self.icons.append(kwargs['icon'])
                else:
                    self.icons = kwargs['icon']

        louie.connect(self.init_complete, 'Coherence.UPnP.Backend.init_completed', louie.Any)
        louie.connect(self.init_failed, 'Coherence.UPnP.Backend.init_failed', louie.Any)
        reactor.callLater(0.2, self.fire, backend, **kwargs)

    def init_failed(self, backend, msg):
        if self.backend != backend:
            return
        self.warning('backend not installed, %s activation aborted - %s' % (self.device_type, msg.getErrorMessage()))
        self.debug(msg)
        try:
            del self.coherence.active_backends[str(self.uuid)]
        except KeyError:
            pass

    def register(self):
        s = self.coherence.ssdp_server
        uuid = str(self.uuid)
        host = self.coherence.hostname
        self.msg('%s register' % self.device_type)
        # we need to do this after the children are there, since we send notifies
        s.register('local',
                    '%s::upnp:rootdevice' % uuid,
                    'upnp:rootdevice',
                    self.coherence.urlbase + uuid[5:] + '/' + 'description-%d.xml' % self.version,
                    host=host)

        s.register('local',
                    uuid,
                    uuid,
                    self.coherence.urlbase + uuid[5:] + '/' + 'description-%d.xml' % self.version,
                    host=host)

        version = self.version
        while version > 0:
            if version == self.version:
                silent = False
            else:
                silent = True
            s.register('local',
                        '%s::urn:schemas-upnp-org:device:%s:%d' % (uuid, self.device_type, version),
                        'urn:schemas-upnp-org:device:%s:%d' % (self.device_type, version),
                        self.coherence.urlbase + uuid[5:] + '/' + 'description-%d.xml' % version,
                        silent=silent,
                        host=host)
            version -= 1


        for service in self._services:
            device_version = self.version
            service_version = self.version
            if hasattr(service, 'version'):
                service_version = service.version
            silent = False

            while service_version > 0:
                try:
                    namespace = service.namespace
                except:
                    namespace = 'schemas-upnp-org'

                device_description_tmpl = 'description-%d.xml' % device_version
                if hasattr(service, 'device_description_tmpl'):
                    device_description_tmpl = service.device_description_tmpl

                s.register('local',
                            '%s::urn:%s:service:%s:%d' % (uuid, namespace, service.id, service_version),
                            'urn:%s:service:%s:%d' % (namespace, service.id, service_version),
                            self.coherence.urlbase + uuid[5:] + '/' + device_description_tmpl,
                            silent=silent,
                            host=host)

                silent = True
                service_version -= 1
                device_version -= 1

    def unregister(self):

        if self.backend != None and hasattr(self.backend, 'release'):
            self.backend.release()

        if not hasattr(self, '_services'):
            """ seems we never made it to actually
                completing that device
            """
            return

        for service in self._services:
            try:
                service.check_subscribers_loop.stop()
            except:
                pass
            if hasattr(service, 'check_moderated_loop') and service.check_moderated_loop != None:
                try:
                    service.check_moderated_loop.stop()
                except:
                    pass
            if hasattr(service, 'release'):
                service.release()
            if hasattr(service, '_release'):
                service._release()

        s = self.coherence.ssdp_server
        uuid = str(self.uuid)
        self.coherence.remove_web_resource(uuid[5:])

        version = self.version
        while version > 0:
            s.doByebye('%s::urn:schemas-upnp-org:device:%s:%d' % (uuid, self.device_type, version))
            for service in self._services:
                if hasattr(service, 'version') and service.version < version:
                    continue
                try:
                    namespace = service.namespace
                except AttributeError:
                    namespace = 'schemas-upnp-org'
                s.doByebye('%s::urn:%s:service:%s:%d' % (uuid, namespace, service.id, version))

            version -= 1

        s.doByebye(uuid)
        s.doByebye('%s::upnp:rootdevice' % uuid)
